import json
import logging
import os
from _decimal import Decimal
from datetime import date, timedelta
from typing import List, Dict, Tuple
import copy

import pandas as pd
from orjson import orjson

from ._typing import (
    DeribitOptionDescription,
    DeribitTokenConfig,
    DeribitMarketStatus,
    OptionMarketBalance,
    OptionPosition,
    OptionKind,
    BuyAction,
    InstrumentStatus,
    Order,
    SellAction,
    ExpiredAction,
    DeliverAction,
    DERIBIT_OPTION_FREQ,
    InsufficientBalanceError,
    DepositAction,
    WithdrawAction,
)
from .helper import round_decimal, position_to_df
from .. import TokenInfo
from .._typing import DemeterError
from ..broker import Market, MarketInfo, write_func, BASE_FREQ
from ..data import CacheManager
from ..utils import (
    float_param_formatter,
    get_formatted_predefined,
    STYLE,
    get_formatted_from_dict,
    console_text,
)

DEFAULT_DATA_PATH = "./data"
BASIC_INTERVAL = pd.Timedelta("1h")


def order_converter(array_str) -> List:
    return json.loads(array_str)


class DeribitOptionMarket(Market):
    """
    The Deribit options market can be utilized for options investment or backtesting of Greek hedging strategies.
    In this market, you can buy or sell options based on the current order book.
    Currently, this market only supports taker orders and does not support maker orders.

    :param market_info: key of this market,
    :type market_info: MarketInfo,
    :param token: token for this market, if you want to trade another token, you can initial another market,
    :type token: TokenInfo,
    :param data: hourly orderbook data of deribit option market for this token
    :type data: pd.DataFrame = None,
    :param data_path: str = path to load data,
    :type data_path: str = DEFAULT_DATA_PATH,

    """

    def __init__(
        self,
        market_info: MarketInfo,
        token: TokenInfo,
        data: pd.DataFrame = None,
        data_path: str = DEFAULT_DATA_PATH,
    ):
        super().__init__(market_info=market_info, data_path=data_path, data=data)
        self.token: TokenInfo = token
        self.token_config: DeribitTokenConfig = DeribitOptionMarket.TOKEN_CONFIGS[token]
        self.positions: Dict[str, OptionPosition] = {}
        self.decimal = self.token_config.min_fee_decimal
        self._balance_cache = None
        # In reality, Deribit is an independent account, and you need to deposit funds into Deribit in order to trade.
        self.balance = Decimal(0)
        self.quote_token = token

    MAX_FEE_RATE = Decimal("0.125")
    ETH = TokenInfo("eth", 18)
    BTC = TokenInfo("btc", 8)
    TOKEN_CONFIGS = {
        ETH: DeribitTokenConfig(
            trade_fee_rate=Decimal("0.0003"),
            delivery_fee_rate=Decimal("0.00015"),
            min_trade_decimal=0,
            min_fee_decimal=-6,
        ),
        BTC: DeribitTokenConfig(
            trade_fee_rate=Decimal("0.0003"),
            delivery_fee_rate=Decimal("0.00015"),
            min_trade_decimal=-1,
            min_fee_decimal=-8,
        ),
    }

    def __str__(self):
        from demeter.utils import orjson_default

        return orjson.dumps(self.description, default=orjson_default).decode()

    @property
    def description(self):
        """
        Get a brief description of this market
        """
        return DeribitOptionDescription(
            type(self).__name__, self._market_info.name, self.token.name, list(self.positions.keys())
        )

    def load_pkl_data(self, path):
        if not os.path.exists(path):
            raise DemeterError(f"{path} doesn't exist")
        self.logger.info(f"start load files in {path}")
        self._data = pd.read_pickle(path)
        self.logger.info("data has been prepared")

    def load_data(self, start_date: date, end_date: date):
        """
        Load data from folder set in data_path. Those data file should be downloaded by demeter, and meet name rule.
        Deribit-option-book-{token}-{day.strftime('%Y%m%d')}.csv
        data can be downloaded from dropbox: https://www.dropbox.com/scl/fo/kwk5kgiseu5rvccjscd0f/ANswtRLzpCxOc6cMTH0oRlE?rlkey=ai071f9695uz287lt8k0bci5e&dl=0

        :param start_date: start day
        :type start_date: date
        :param end_date: end day, the end day will be included
        :type end_date: date
        """

        cache_key = CacheManager.get_cache_key(self.market_info.type.name, start_date, end_date)
        cache_df = CacheManager.load(cache_key)
        if cache_df is not None:
            self.data = cache_df
            return

        self.logger.info(f"start load files from {start_date} to {end_date}...")
        day = start_date
        df = pd.DataFrame()
        from tqdm import tqdm

        with tqdm(total=(end_date - start_date).days + 1, ncols=150) as pbar:
            while day <= end_date:
                path = os.path.join(
                    self.data_path,
                    f"Deribit-option-book-{self.token.name}-{day.strftime('%Y%m%d')}.csv",
                )
                if not os.path.exists(path):
                    logging.warning(f"resource file {path} not found")
                    day += timedelta(days=1)
                    pbar.update()
                    continue

                day_df = pd.read_csv(
                    str(path),
                    parse_dates=["time", "expiry_time"],
                    index_col=["time", "instrument_name"],
                    converters={"asks": order_converter, "bids": order_converter},
                )
                day_df.drop(columns=["actual_time", "min_price", "max_price"], inplace=True)
                df = pd.concat([df, day_df])
                day += timedelta(days=1)
                pbar.update()

        self._data = df
        CacheManager.save(cache_key, df)
        self.logger.info("data has been prepared")

    @float_param_formatter
    def deposit(self, amount: Decimal | float) -> Decimal:
        """
        Add token amount to wallet of deribit, token is quote token in this market.
        :param amount: balance, e.g. 1.2345
        :type amount: Decimal | float
        :return: new balance
        :rtype: Decimal
        """
        self.broker.subtract_from_balance(self.token, amount)
        self._add_to_balance(amount)
        self._record_action(
            DepositAction(
                market=self._market_info,
                token=self.token.name,
                amount=amount,
            )
        )

    def _add_to_balance(self, amount: Decimal | float) -> Decimal:
        self.balance += amount
        return self.balance

    @float_param_formatter
    def withdraw(self, amount: Decimal | float) -> Decimal:
        """
        | Sub token amount to wallet of deribit, token is quote token in this market.

        :param amount: Decimal or float type
        :type amount:  Decimal | float
        :return: new balance
        :rtype: Decimal
        """
        new_balance = self._subtract_from_balance(amount)
        # Actually, withdraw fee should be charged, but the amount is depended on network condition
        # https://www.deribit.com/kb/fees
        self.broker.add_to_balance(self.token, amount)
        self._record_action(
            WithdrawAction(
                market=self._market_info,
                token=self.token.name,
                amount=amount,
            )
        )
        return new_balance

    def _subtract_from_balance(self, amount: Decimal | float) -> Decimal:
        left_over = self.balance - amount
        if left_over < Decimal(0):
            raise InsufficientBalanceError(f"Not enough balance, balance is {self.balance}, amount to sub is {amount}")
        self.balance = left_over
        return left_over

    def check_market(self):
        """
        check market before back test
        """
        if not isinstance(self.data, pd.DataFrame):
            raise DemeterError("data must be type of data frame")

    def set_market_status(
        self,
        data: DeribitMarketStatus,
        price: pd.Series,
    ):
        """
        Set up market status, such as liquidity, price

        :param data: market status
        :type data: Series | DeribitMarketStatus
        :param price: current price of all relative token, based in usd
        :type price: Series

        """
        super().set_market_status(data, price)
        if data.data is None:
            tmr_idx = data.timestamp.floor(DERIBIT_OPTION_FREQ)
            if tmr_idx in self._data.index:
                data.data = self._data.loc[tmr_idx]
            else:
                data.data = pd.DataFrame(columns=self._data.columns)
                if self._is_open():
                    logging.warning(f"Deribit data in {data.timestamp} doesn't exist")
        self._market_status = data

    # region for option market only

    def get_trade_fee(self, amount: Decimal, total_premium: Decimal) -> Decimal:
        """
        Calculate trade fee, according to https://www.deribit.com/kb/fees

        :param amount: instrument amount
        :type amount: Decimal
        :param total_premium: total value of instruments
        :type total_premium: Decimal
        """
        return round_decimal(
            min(
                self.token_config.trade_fee_rate * amount,
                DeribitOptionMarket.MAX_FEE_RATE * total_premium,
            ),
            self.decimal,
        )

    def get_deliver_fee(self, amount: Decimal, total_premium: Decimal):
        """
        Calculate fee of deliver when exercising, according to https://www.deribit.com/kb/fees

        :param amount: instrument amount
        :type amount: Decimal
        :param total_premium: total value of instruments
        :type total_premium: Decimal
        """
        return round_decimal(
            min(
                self.token_config.delivery_fee_rate * amount,
                DeribitOptionMarket.MAX_FEE_RATE * total_premium,
            ),
            self.decimal,
        )

    def get_price_from_data(self) -> pd.Series:
        """
        Get hourly underlying price.
        """
        if self._data is None:
            raise DemeterError("data is empty")
        price = []
        for hour, hour_df in self._data.groupby(level=0):
            price.append({"time": hour, self.token.name: hour_df.iloc[0]["underlying_price"]})
        price_df = pd.DataFrame(price)
        price_df.set_index(["time"], inplace=True)
        # expend to the end of the day
        price_df.loc[price_df.tail(1).index[0].ceil("1d")] = 0
        price_df = price_df.resample(BASE_FREQ).ffill()
        return price_df.drop(price_df.index[-1])

    def __get_trade_amount(self, amount: Decimal):
        """
        Round trade amount, and ensure amount is above dust amount

        :param amount: option amount to buy/sell
        :param amount: Decimal
        """
        if amount < self.token_config.min_amount:
            return self.token_config.min_amount
        return round_decimal(amount, self.token_config.min_trade_decimal)

    def formatted_str(self):
        """
        Return a brief description of this market in pretty format. Used for print in console.
        """
        value = get_formatted_predefined(f"{self.market_info.name}({type(self).__name__})", STYLE["header3"]) + "\n"
        token_dict = {"token": self.token.name}
        value += get_formatted_from_dict(token_dict) + "\n"
        balance: OptionMarketBalance = self.get_market_balance()
        value += (
            get_formatted_from_dict(
                {
                    "delta": console_text.format_value(balance.delta),
                    "gamma": console_text.format_value(balance.gamma),
                }
            )
            + "\n"
        )
        value += get_formatted_predefined("Positions", STYLE["key"]) + "\n"
        supply_df = position_to_df(self.positions)
        value += supply_df.to_string() + "\n" if len(supply_df.index) > 0 else "Empty DataFrame\n"

        return value

    @float_param_formatter
    def estimate_cost(
        self,
        instrument_name: str,
        amount: float | Decimal,
        type: str = "buy",  # buy or sell
        price_in_token: float | Decimal | None = None,
    ):
        amount = self.__get_trade_amount(amount)
        row = self.data.loc[(self._market_status.timestamp, instrument_name)]
        order_list = row.asks if type == "buy" else row.bids
        order_list = copy.deepcopy(order_list)
        used_order = self._deduct_order_amount(amount, order_list, price_in_token)

        total_premium = Decimal(sum([Decimal(t.amount) * Decimal(t.price) for t in used_order]))
        fee_amount = self.get_trade_fee(amount, total_premium)
        return total_premium + fee_amount

    @write_func
    @float_param_formatter
    def buy(
        self,
        instrument_name: str,
        amount: float | Decimal,
        price_in_token: float | Decimal | None = None,
        price_in_usd: float | Decimal | None = None,
        max_mark_price_multiple: float | Decimal | None = None,
    ) -> Tuple[List[Order], Decimal]:
        """
        Buy option.
        if price is not none, will buy at specific price
        or else will buy according to bids

        :param instrument_name: instrument name
        :type instrument_name: str,
        :param amount: amount to buy,
        :type amount: float | Decimal,
        :param price_in_token: price, based in token,
        :type price_in_token: float | Decimal | None = None,
        :param price_in_usd: price, based in usd,
        :type price_in_usd: float | Decimal | None = None,
        :param max_mark_price_multiple: times to mark_price, if order price is greater than mark_price * max_allowed, will not buy at this price


        """
        amount, instrument, price_in_token = self._check_transaction(
            instrument_name, amount, price_in_token, price_in_usd, True, max_mark_price_multiple
        )

        # this actually pass reference of the array, so asks array will be updated in _deduct_order_amount.
        # so when orders are deducted, asks order number will be changed.
        if max_mark_price_multiple is not None:
            asks = list(
                filter(lambda x: x[0] < max_mark_price_multiple * Decimal(instrument.mark_price), instrument.asks)
            )
        else:
            asks = instrument.asks
        # deduct bids amount
        ask_list = self._deduct_order_amount(amount, asks, price_in_token)

        total_premium = Decimal(sum([Decimal(t.amount) * Decimal(t.price) for t in ask_list]))
        fee_amount = self.get_trade_fee(amount, total_premium)
        self._subtract_from_balance(total_premium + fee_amount)

        # add position
        average_price = Order.get_average_price(ask_list)
        if instrument_name not in self.positions.keys():
            self.positions[instrument_name] = OptionPosition(
                instrument_name=instrument_name,
                expiry_time=instrument.expiry_time,
                strike_price=instrument.strike_price,
                type=OptionKind(instrument.type),
                amount=amount,
                avg_buy_price=average_price,
                buy_amount=amount,
                avg_sell_price=Decimal(0),
                sell_amount=Decimal(0),
            )
        else:
            position = self.positions[instrument_name]
            position.avg_buy_price = Order.get_average_price(
                [
                    Order(average_price, amount),
                    Order(position.avg_buy_price, position.buy_amount),
                ]
            )
            position.buy_amount += amount
            position.amount += amount
        self._record_action(
            BuyAction(
                market=self._market_info,
                instrument_name=instrument_name,
                type=OptionKind(instrument.type),
                average_price=average_price,
                amount=amount,
                total_premium=total_premium,
                mark_price=Decimal(str(instrument.mark_price)),
                underlying_price=Decimal(str(instrument.underlying_price)),
                fee=fee_amount,
                orders=ask_list,
            )
        )
        return ask_list, fee_amount

    @staticmethod
    def _find_available_orders(price, order_list) -> List:
        error = Decimal("0.001")
        return list(filter(lambda x: (1 - error) * price < x[0] < (1 + error) * price, order_list))

    @write_func
    @float_param_formatter
    def sell(
        self,
        instrument_name: str,
        amount: float | Decimal,
        price_in_token: float | Decimal | None = None,
        price_in_usd: float | Decimal | None = None,
        max_mark_price_multiple: float | Decimal | None = None,
    ) -> Tuple[List[Order], Decimal]:
        """
        Sell option.
        if price is not none, will sell at specific price
        or else will sell according to asks

        :param instrument_name: instrument name
        :type instrument_name: str,
        :param amount: amount to buy,
        :type amount: float | Decimal,
        :param price_in_token: price, based in token,
        :type price_in_token: float | Decimal | None = None,
        :param price_in_usd: price, based in usd,
        :type price_in_usd: float | Decimal | None = None,
        :param max_mark_price_multiple: times to mark_price, if order price is greater than mark_price * max_allowed, will not buy at this price

        """
        amount, instrument, price_in_token = self._check_transaction(
            instrument_name, amount, price_in_token, price_in_usd, False, max_mark_price_multiple
        )

        # deduct  amount
        if max_mark_price_multiple is not None:
            bids = list(
                filter(lambda x: x[0] > max_mark_price_multiple / Decimal(instrument.mark_price), instrument.bids)
            )
        else:
            bids = instrument.bids

        bid_list = self._deduct_order_amount(amount, bids, price_in_token)

        # write positions back
        # if self.data is not None:
        #     self.data.loc[(self._market_status.timestamp, instrument_name), "bids"] = bids

        total_premium = Decimal(sum([Decimal(t.amount) * Decimal(t.price) for t in bid_list]))
        fee = self.get_trade_fee(amount, total_premium)
        self._add_to_balance(total_premium - fee)

        # subtract position
        average_price = Order.get_average_price(bid_list)
        if instrument_name not in self.positions.keys():
            raise DemeterError("No such instrument position")

        position = self.positions[instrument_name]
        position.avg_sell_price = Order.get_average_price(
            [
                Order(average_price, amount),
                Order(position.avg_sell_price, position.sell_amount),
            ]
        )
        position.sell_amount += amount
        position.amount -= amount

        if position.amount <= Decimal(0):
            del self.positions[instrument_name]

        self._record_action(
            SellAction(
                market=self._market_info,
                instrument_name=instrument_name,
                type=OptionKind(instrument.type),
                average_price=average_price,
                amount=amount,
                total_premium=total_premium,
                mark_price=Decimal(str(instrument.mark_price)),
                underlying_price=Decimal(str(instrument.underlying_price)),
                fee=fee,
                orders=bid_list,
            )
        )
        return bid_list, fee

    def _deduct_order_amount(self, amount, orders, price_in_token) -> List[Order]:
        """
        subtract amount from asks/bids. e.g. if bid1 is run out, will deduct bid2. etc.
        """
        order_list = []
        if price_in_token is not None:
            for order in orders:
                if price_in_token == Decimal(str(order[0])):
                    order[1] -= amount
                    order_list.append(Order(price_in_token, amount))
        else:
            amount_to_deduct = amount
            for order in orders:
                if order[1] == 0 or order[1] == Decimal(0):
                    continue
                should_deduct = min(Decimal(str(order[1])), amount_to_deduct)
                amount_to_deduct -= should_deduct
                order[1] -= float(should_deduct)
                order_list.append(Order(Decimal(str(order[0])), should_deduct))
                if order[1] > 0 or amount_to_deduct == Decimal(0):
                    break
        return order_list

    def _check_transaction(
        self, instrument_name, amount, price_in_token, price_in_usd, is_buy, max_mark_price_multiple=None
    ) -> Tuple[Decimal, InstrumentStatus, Decimal]:
        """
        | Check transaction,
        | - ensure instrument exists
        | - ensure instrument is open
        | - ensure there are enough amount
        | - ensure price is in asks/bids

        :param max_mark_price_multiple: times to mark_price, if order price is greater than mark_price * max_allowed, will not buy beyond price
        """
        if instrument_name not in self._market_status.data.index:
            raise DemeterError(f"{instrument_name} is not in current orderbook")
        instrument: InstrumentStatus = self._market_status.data.loc[instrument_name]
        if instrument.state != "open":
            raise DemeterError(f"state of {instrument_name} is not open")
        if amount < self.token_config.min_amount:
            raise DemeterError(
                f"amount should greater than min amount {self.token_config.min_amount} {self.token.name}"
            )
        amount = self.__get_trade_amount(amount)
        if price_in_usd is not None and price_in_token is None:
            price_in_token = price_in_usd / instrument.underlying_price

        if is_buy:
            if max_mark_price_multiple is not None:
                max_price = max_mark_price_multiple * Decimal(instrument.mark_price)
                available_orders = list(filter(lambda x: x[0] < max_price, instrument.asks))
            else:
                available_orders = instrument.asks
        else:
            if max_mark_price_multiple is not None:
                min_price = max_mark_price_multiple / Decimal(instrument.mark_price)
                available_orders = list(filter(lambda x: x[0] > min_price, instrument.bids))
            else:
                available_orders = instrument.bids

        if price_in_token is not None:
            # to prevent error in decimal
            available_orders = DeribitOptionMarket._find_available_orders(price_in_token, available_orders)

            if len(available_orders) < 1:
                raise DemeterError(
                    f"{instrument_name} doesn't have a order in price {price_in_token} {self.token.name}"
                )
            price_in_token = Decimal(str(available_orders[0][0]))
            available_amount = Decimal(available_orders[0][1])
        else:
            available_amount = sum([Decimal(x[1]) for x in available_orders])
        if amount > available_amount:
            raise DemeterError(
                f"insufficient order to buy {instrument_name}, required amount is {amount}, "
                f"available amount is {available_amount}"
            )

        return amount, instrument, price_in_token

    def update(self):
        """
        Trigger update of this market
        """
        if self._is_open():
            self.check_option_exercise()

    def _is_open(self):
        """
        ensure this market is writable. e.g. deribit option market only has data at the hour,
        but uniswap data is minutely. so at the rest 59 minutes, deribit option market is readonly.
        which means, you can read status, but you can not buy or sell options.
        """
        return self._market_status.timestamp == self._market_status.timestamp.floor(DERIBIT_OPTION_FREQ)

    def get_market_balance(self) -> OptionMarketBalance:
        """
        Get market asset balance, such as current positions, net values

        :return: Balance in this market includes net value, position value
        :rtype: MarketBalance
        """
        if self._is_open():
            total_premium = Decimal(0)
            delta = gamma = Decimal(0)
            for position in self.positions.values():
                # data may missing due to unstable collect server
                if position.instrument_name not in self.market_status.data.index:
                    continue
                instr_status = self.market_status.data.loc[position.instrument_name]
                instrument_premium = position.amount * round_decimal(instr_status.mark_price, self.decimal)
                total_premium += instrument_premium
                delta += instrument_premium * round_decimal(instr_status.delta, self.decimal)
                gamma += instrument_premium * round_decimal(instr_status.gamma, self.decimal)

            delta = Decimal(0) if total_premium == Decimal(0) else delta / total_premium
            gamma = Decimal(0) if total_premium == Decimal(0) else gamma / total_premium
            equity = self.balance + total_premium
            self._balance_cache = OptionMarketBalance(equity, self.balance, total_premium, delta, gamma)
        return self._balance_cache

    # region exercise
    def check_option_exercise(self):
        """
        | loop all the option position,
        | if expired, if option position is in the money, then exercise.
        | if out of the money, then abandon
        """
        key_to_remove = []
        for pos_key, position in self.positions.items():
            if self._market_status.timestamp >= position.expiry_time:
                # should not happen
                if position.instrument_name in self._market_status.data.index:
                    instrument: InstrumentStatus = self.market_status.data.loc[position.instrument_name]
                else:
                    logging.warning(f"{position.instrument_name} is not in current orderbook")
                    instrument = InstrumentStatus(mark_price=0, underlying_price=self._price_status[self.token.name])

                deliver_amount = deliver_fee = None
                if position.type == OptionKind.put and position.strike_price > instrument.underlying_price:
                    deliver_amount, deliver_fee = self._deliver_option(position, instrument, False)
                elif position.type == OptionKind.call and position.strike_price < instrument.underlying_price:
                    deliver_amount, deliver_fee = self._deliver_option(position, instrument, True)
                if deliver_amount is not None:
                    self._record_action(
                        DeliverAction(
                            market=self._market_info,
                            instrument_name=pos_key,
                            type=position.type,
                            mark_price=round_decimal(instrument.mark_price, self.decimal),
                            amount=position.amount,
                            total_premium=position.amount * round_decimal(instrument.mark_price, self.decimal),
                            strike_price=position.strike_price,
                            underlying_price=round_decimal(instrument.underlying_price, self.decimal),
                            deriver_amount=deliver_amount,
                            fee=deliver_fee,
                            income_amount=deliver_amount - deliver_fee,
                        )
                    )
                key_to_remove.append(pos_key)

        for pos_key in key_to_remove:
            position = self.positions[pos_key]
            if pos_key in self.market_status.data.index:
                instrument: InstrumentStatus = self.market_status.data.loc[position.instrument_name]
            else:
                instrument = InstrumentStatus(mark_price=0, underlying_price=self._price_status[self.token.name])
            self._record_action(
                ExpiredAction(
                    market=self._market_info,
                    instrument_name=pos_key,
                    type=position.type,
                    mark_price=round_decimal(instrument.mark_price, self.decimal),
                    amount=position.amount,
                    total_premium=position.amount * round_decimal(instrument.mark_price, self.decimal),
                    strike_price=position.strike_price,
                    underlying_price=round_decimal(instrument.underlying_price, self.decimal),
                )
            )
            del self.positions[pos_key]

    def _deliver_option(
        self, option_pos, instrument: InstrumentStatus, is_call
    ) -> Tuple[Decimal | None, Decimal | None]:
        """
        deliver option
        """
        fee = self.get_deliver_fee(
            option_pos.amount,
            option_pos.amount * round_decimal(instrument.mark_price, self.decimal),
        )
        if is_call:
            price_diff = instrument.underlying_price - option_pos.strike_price
        else:
            price_diff = option_pos.strike_price - instrument.underlying_price

        balance_to_add = round_decimal(
            option_pos.amount * Decimal(price_diff / instrument.underlying_price),
            self.decimal,
        )
        if balance_to_add <= fee:
            return None, None
        self._add_to_balance(balance_to_add - fee)
        return balance_to_add, fee

    # endregion

    # endregion

    def _resample(self, freq: str):
        interval_delta = pd.Timedelta(freq)
        if interval_delta <= BASIC_INTERVAL:
            return
        else:
            self._data = self._data.groupby(level=1).resample(freq, level=0).first().swaplevel(1, 0)



import numpy as np
import os
import pandas as pd
from datetime import date, timedelta, datetime, time
from decimal import Decimal
from typing import Dict, Tuple
import orjson

from ._typing import (
    UniV3Pool,
    TokenInfo,
    Position,
    UniLpBalance,
    AddLiquidityAction,
    RemoveLiquidityAction,
    CollectFeeAction,
    BuyAction,
    SellAction,
    position_dict_to_dataframe,
    PositionInfo,
    UniDescription,
    UniswapMarketStatus,
    SwapAction,
    PositionStatus,
)
from .core import V3CoreLib
from .data import fillna, resample
from .helper import (
    tick_to_base_unit_price,
    base_unit_price_to_tick,
    base_unit_price_to_sqrt_price_x96,
    sqrt_price_x96_to_base_unit_price,
    tick_to_sqrt_price_x96,
    get_swap_value_with_part_balance_used,
    MIN_ERROR,
    nearest_usable_tick,
    sqrt_price_x96_to_tick,
)
from .liquitidy_math import (
    get_sqrt_ratio_at_tick,
    estimate_ratio,
    get_liquidity,
    get_liquidity_for_amount0,
    get_liquidity_for_amount1,
)
from .. import MarketTypeEnum
from .._typing import DemeterError, DECIMAL_0, UnitDecimal
from ..broker import MarketBalance, Market, MarketInfo, write_func
from ..data import CacheManager
from ..utils import (
    get_formatted_from_dict,
    get_formatted_predefined,
    STYLE,
    float_param_formatter,
    to_decimal,
    require,
)


class UniLpMarket(Market):
    """
    | UniLpMarket is the simulator of uniswap v3, it can simulate transactions such as add/remove liquidity, swap assets. and calculate position net value
    | UniLpMarket corresponds to a pool on chain, which means a token pair in a chain.

    :param market_info: key of this market
    :type market_info: MarketInfo
    :param pool_info: Uniswap v3 pool info
    :type pool_info: UniV3Pool
    :param data: pool data for back test. downloaded by demeter-fetch
    :type data: pd.DataFrame
    :param data_path: path to load pool data
    :type data_path: str
    """

    def __init__(
        self, market_info: MarketInfo, pool_info: UniV3Pool, data: pd.DataFrame = None, data_path: str = "./data"
    ):
        super().__init__(market_info=market_info, data=data, data_path=data_path)
        self._pool: UniV3Pool = pool_info
        # init balance
        self._is_token0_quote = pool_info.is_token0_quote
        # reference for super().assets dict.
        self.base_token, self.quote_token = self._convert_pair(self.pool_info.token0, self.pool_info.token1)
        # status
        self._positions: Dict[PositionInfo, Position] = {}
        # In order to distinguish price in pool and to u, we call former one "pool price"
        self._pool_price_unit = f"{self.base_token.name}/{self.quote_token.name}"
        # internal temporary variable
        # self.action_buffer = []
        # tick of last minute(previous minute), to compatible with old version, keep default as None
        self.last_tick: int = None

    # region properties

    def __str__(self):
        from demeter.utils import orjson_default

        return orjson.dumps(self.description, default=orjson_default).decode()

    @property
    def description(self) -> UniDescription:
        return UniDescription(
            type=type(self).__name__,
            name=self._market_info.name,
            token0=self.pool_info.token0,
            token1=self.pool_info.token1,
            quote_token=self.pool_info.quote_token,
            base_token=(
                self.pool_info.token0 if self.pool_info.quote_token == self.pool_info.token1 else self.pool_info.token1
            ),
            fee_rate=self.pool_info.fee_rate,
        )

    @property
    def positions(self) -> Dict[PositionInfo, Position]:
        """
        current positions in broker

        :return: all positions
        :rtype: dict[PositionInfo:Position]
        """
        return self._positions

    @property
    def pool_info(self) -> UniV3Pool:
        """
        Get pool info.

        :return: pool info
        :rtype: UniV3Pool
        """
        return self._pool

    @property
    def token0(self) -> TokenInfo:
        """
        get asset 0 info, including balance

        :return: BrokerAsset
        :rtype: BrokerAsset
        """
        return self._pool.token0

    @property
    def token1(self) -> TokenInfo:
        """
        get asset 1 info, including balance

        :return: BrokerAsset
        :rtype: BrokerAsset
        """
        return self._pool.token1

    @property
    def market_status(self) -> UniswapMarketStatus:
        return self._market_status

    # endregion

    def get_position(self, position_info: PositionInfo) -> Position:
        """
        get get_position by get_position information

        :param position_info: get_position information
        :type position_info: PositionInfo
        :return: Position entity
        :rtype: Position
        """
        return self._positions[position_info]

    def set_market_status(
        self,
        market_status: UniswapMarketStatus | None,
        price: pd.Series | None,
    ):
        """
        Set current pool status (total liquidity, price etc.) to Market

        | Note: UniswapMarketStatus also have price attribute, but it's different price parameter.
        | price parameter is used to evaluate token value. Often get from coingecko
        | UniswapMarketStatus.price is the relative price of pool token pair. It's calculated from pool swap events.
        | if base token in pool is stable coin and its price is 1, the two price will be the same. But if stable coin price is not 1, there will be a gap between two prices.

        :param market_status: market data
        :type market_status: UniswapMarketStatus
        :param price: price of token at this moment
        :type price: pd.Series
        """
        # update price tick
        super().set_market_status(market_status, price)

        total_virtual_liq = sum([p.liquidity for p in self._positions.values()])
        self.last_tick = self._market_status.data.closeTick if "closeTick" in self._market_status.data.index else np.nan

        if market_status.data is None:
            market_status.data = self.data.loc[market_status.timestamp].copy()
        market_status.data.currentLiquidity = market_status.data.currentLiquidity + total_virtual_liq
        self._market_status = market_status

    def get_price_from_data(self) -> Tuple[pd.DataFrame, TokenInfo]:
        """
        Extract token pair price from pool data.

        :return: a dataframe includes quote token price, and quote token price will be set to 1
        :rtype: Tuple[pd.DataFrame, TokenInfo]

        """
        if self.data is None:
            raise DemeterError("data has not set")
        price_series: pd.Series = self.data.price
        df = pd.DataFrame(index=price_series.index, data={self.base_token.name: price_series})
        df[self.quote_token.name] = 1
        return df, self.quote_token

    def _convert_pair(self, any0, any1):
        """
        convert order of token0/token1 to base_token/quote_token, according to self.is_token0_quote.

        Or convert order of base_token/quote_token to token0/token1

        :param any0: token0 or any property of token0, e.g. balance...
        :param any1: token1 or any property of token1, e.g. balance...
        :return: (base,qoute) or (token0,token1)
        """
        return (any1, any0) if self._is_token0_quote else (any0, any1)

    def check_market(self):
        """
        Verify settings before backtest
        """
        super().check_market()
        required_columns = [
            "closeTick",
            "currentLiquidity",
            "inAmount0",
            "inAmount1",
            "price",
        ]
        for col in required_columns:
            assert col in self.data.columns

        if not self._pool:
            raise DemeterError("set up pool info first")
        if self.base_token not in self.broker.assets:
            self.broker.set_balance(self.base_token, DECIMAL_0)
        if self.quote_token not in self.broker.assets:
            self.broker.set_balance(self.quote_token, DECIMAL_0)

    def update(self):
        """
        re-calculate status.
        """
        self.__update_fee()

    def __update_fee(self):
        """
        update fee in all positions according to current status

        fee will be calculated by liquidity
        """
        for position_info, position in self._positions.items():
            V3CoreLib.update_fee(self.last_tick, self.pool_info, position_info, position, self.market_status.data)

    def _get_value(self, amount0, amount1, pool_price):
        base, quote = self._convert_pair(amount0, amount1)
        return base * pool_price + quote

    def get_position_status(self, pos_key: PositionInfo) -> PositionStatus:
        require(pos_key in self.positions, "Position not exist")

        pool_price = self._market_status.data.price
        liq_amount0, liq_amount1 = self.get_position_amount(pos_key)
        amount0 = liq_amount0 + self.positions[pos_key].pending_amount0
        amount1 = liq_amount1 + self.positions[pos_key].pending_amount1
        return PositionStatus(
            liquidity=self.positions[pos_key].liquidity,
            liquidity_amount0=liq_amount0,
            liquidity_amount1=liq_amount1,
            liquidity_value=self._get_value(liq_amount0, liq_amount1, pool_price),
            pending_amount0=self.positions[pos_key].pending_amount0,
            pending_amount1=self.positions[pos_key].pending_amount1,
            pending_value=self._get_value(
                self.positions[pos_key].pending_amount0, self.positions[pos_key].pending_amount1, pool_price
            ),
            amount0=amount0,
            amount1=amount1,
            value=self._get_value(amount0, amount1, pool_price),
            H=self.positions[pos_key].upper_price / self.positions[pos_key].init_price,
            L=self.positions[pos_key].lower_price / self.positions[pos_key].init_price,
            P=pool_price / self.positions[pos_key].init_price,
        )

    def get_position_amount(self, position_info: PositionInfo) -> Tuple[Decimal, Decimal]:
        if position_info not in self.positions:
            return DECIMAL_0, DECIMAL_0
        pool_price = self._market_status.data.price
        sqrt_price = base_unit_price_to_sqrt_price_x96(
            pool_price,
            self._pool.token0.decimal,
            self._pool.token1.decimal,
            self._is_token0_quote,
        )
        position = self.positions[position_info]
        amount0, amount1 = V3CoreLib.get_token_amounts(self._pool, position_info, sqrt_price, position.liquidity)
        return amount0, amount1

    def get_market_balance(self) -> MarketBalance:
        """
        get current status, including positions, balances

        :return: MarketBalance
        """
        pool_price = {self.quote_token.name: Decimal(1), self.base_token.name: self._market_status.data.price}

        sqrt_price = base_unit_price_to_sqrt_price_x96(
            self._market_status.data.price,
            self._pool.token0.decimal,
            self._pool.token1.decimal,
            self._is_token0_quote,
        )
        base_fee_sum = Decimal(0)
        quote_fee_sum = Decimal(0)
        deposit_amount0 = Decimal(0)
        deposit_amount1 = Decimal(0)
        for position_info, position in self._positions.items():
            if position.transferred:
                continue
            base_fee, quote_fee = self._convert_pair(position.pending_amount0, position.pending_amount1)
            base_fee_sum += base_fee
            quote_fee_sum += quote_fee
            amount0, amount1 = V3CoreLib.get_token_amounts(self._pool, position_info, sqrt_price, position.liquidity)
            deposit_amount0 += amount0
            deposit_amount1 += amount1

        base_deposit_amount, quote_deposit_amount = self._convert_pair(deposit_amount0, deposit_amount1)

        net_value = (base_fee_sum + base_deposit_amount) * pool_price[self.base_token.name] + (
            quote_fee_sum + quote_deposit_amount
        ) * pool_price[self.quote_token.name]

        val = UniLpBalance(
            net_value=net_value,
            base_uncollected=UnitDecimal(base_fee_sum, self.base_token.name),
            quote_uncollected=UnitDecimal(quote_fee_sum, self.quote_token.name),
            base_in_position=UnitDecimal(base_deposit_amount, self.base_token.name),
            quote_in_position=UnitDecimal(quote_deposit_amount, self.quote_token.name),
            position_count=len(list(filter(lambda p: not p.transferred, self._positions.values()))),
        )
        return val

    def transfer_position_out(self, position_info: PositionInfo):
        if position_info in self.positions and not self.positions[position_info].transferred:
            self.positions[position_info].transferred = True
        else:
            raise DemeterError("position not exist or has transferred out ")

    def transfer_position_in(self, position_info: PositionInfo):
        if position_info in self.positions and self.positions[position_info].transferred:
            self.positions[position_info].transferred = False
        else:
            raise DemeterError("position not exist or has not transferred yet ")

    def tick_to_price(self, tick: int) -> Decimal:
        """
        convert tick to price

        :param tick: tick
        :type tick: int
        :return: price
        :rtype: Decimal
        """
        return tick_to_base_unit_price(
            int(tick),
            self._pool.token0.decimal,
            self._pool.token1.decimal,
            self._is_token0_quote,
        )

    def estimate_liquidity(self, value: Decimal, position: PositionInfo) -> Tuple[int, Decimal, Decimal]:
        """
        In the current market situation, to obtain value (e.g. 500 usd) from this tick range, how many liquidity units need to be removed?

        :param value: value to get (e.g. 500 usd)
        :param position: LP position
        :return: Liquidity and how much token amount you can get if remove liquidity.
        """
        sqrt_price_x96 = base_unit_price_to_sqrt_price_x96(
            self.market_status.data.price,
            self.pool_info.token0.decimal,
            self.pool_info.token1.decimal,
            self.pool_info.is_token0_quote,
        )
        current_tick = sqrt_price_x96_to_tick(sqrt_price_x96)

        lower_sqrt = get_sqrt_ratio_at_tick(position.lower_tick)
        upper_sqrt = get_sqrt_ratio_at_tick(position.upper_tick)
        if current_tick <= position.lower_tick:
            quote_amount = DECIMAL_0
            base_amount = value / Decimal(self.market_status.data.price)
            liq = int(get_liquidity_for_amount0(lower_sqrt, upper_sqrt, base_amount * 10**self.base_token.decimal))
            token0_amount, token1_amount = self._convert_pair(base_amount, quote_amount)

        elif current_tick >= position.upper_tick:
            quote_amount = value
            base_amount = DECIMAL_0
            liq = int(get_liquidity_for_amount1(lower_sqrt, upper_sqrt, quote_amount * 10**self.quote_token.decimal))
            token0_amount, token1_amount = self._convert_pair(base_amount, quote_amount)
        else:
            token0_amount, token1_amount = self.estimate_amount(value, position.lower_tick, position.upper_tick)
            liq = get_liquidity(
                sqrt_price_x96,
                position.lower_tick,
                position.upper_tick,
                token0_amount,
                token1_amount,
                self.pool_info.token0.decimal,
                self.pool_info.token1.decimal,
            )

        return liq, token0_amount, token1_amount

    def estimate_amount(self, value: Decimal, lower_tick: int, upper_tick: int) -> Tuple[Decimal, Decimal]:
        """
        Emulate amount of token 0,1 with given tick range

        :param value: value to get (e.g. 500 usd)
        :param lower_tick: lower tick
        :param upper_tick: upper tick
        :return:
        """
        current_price = self.market_status.data.price
        current_tick = base_unit_price_to_tick(
            current_price, self.pool_info.token0.decimal, self.pool_info.token1.decimal, self.pool_info.is_token0_quote
        )
        ratio_in_amount = estimate_ratio(current_tick, lower_tick, upper_tick)
        ratio_in_amount = Decimal(
            ratio_in_amount * 10 ** (self.pool_info.token1.decimal - self.pool_info.token0.decimal)
        )
        ratio_in_value = (
            ratio_in_amount / current_price if self.pool_info.is_token0_quote else ratio_in_amount * current_price
        )
        token1_value = value / (ratio_in_value + 1)
        token0_value = value - token1_value
        token0_amount = token0_value if self.pool_info.is_token0_quote else token0_value / current_price
        token1_amount = token1_value if not self.pool_info.is_token0_quote else token1_value / current_price
        return token0_amount, token1_amount

    @float_param_formatter
    def price_to_tick(self, price: Decimal | float) -> int:
        """
        convert price to tick

        :param price: price
        :type price:  Decimal | float
        :return: tick
        :rtype: int
        """
        return nearest_usable_tick(
            base_unit_price_to_tick(
                price,
                self._pool.token0.decimal,
                self._pool.token1.decimal,
                self._is_token0_quote,
            ),
            self.pool_info.tick_spacing,
        )

    @write_func
    def _add_liquidity_by_tick(
        self,
        token0_amount: Decimal,
        token1_amount: Decimal,
        lower_tick: int,
        upper_tick: int,
        sqrt_price_x96: int = -1,
    ):
        lower_tick = int(lower_tick)
        upper_tick = int(upper_tick)
        require(
            lower_tick % self.pool_info.tick_spacing == 0 and upper_tick % self.pool_info.tick_spacing == 0,
            "tick should match tick space",
        )
        sqrt_price_x96 = int(sqrt_price_x96)

        if sqrt_price_x96 == -1:
            # self.current_tick must be initialed
            sqrt_price_x96 = get_sqrt_ratio_at_tick(self.market_status.data.closeTick)
        if lower_tick > upper_tick:
            raise DemeterError("lower tick should be less than upper tick")

        token0_used, token1_used, liquidity, position_info = V3CoreLib.new_position(
            self._pool,
            token0_amount,
            token1_amount,
            lower_tick,
            upper_tick,
            sqrt_price_x96,
        )
        if position_info in self._positions:
            self._positions[position_info].liquidity += liquidity
        else:
            lower_price = self.tick_to_price(lower_tick)
            upper_price = self.tick_to_price(upper_tick)
            init_price = sqrt_price_x96_to_base_unit_price(
                sqrt_price_x96,
                self.pool_info.token0.decimal,
                self.pool_info.token1.decimal,
                self.pool_info.is_token0_quote,
            )
            if self.pool_info.is_token0_quote:
                lower_price, upper_price = upper_price, lower_price
            self._positions[position_info] = Position(
                DECIMAL_0, DECIMAL_0, liquidity, lower_price, upper_price, init_price
            )
        self.broker.subtract_from_balance(self.token0, token0_used)
        self.broker.subtract_from_balance(self.token1, token1_used)
        return position_info, token0_used, token1_used, liquidity

    @write_func
    def __remove_liquidity(self, position: PositionInfo, liquidity: int = None, sqrt_price_x96: int = -1):
        sqrt_price_x96 = (
            int(sqrt_price_x96)
            if sqrt_price_x96 != -1
            else base_unit_price_to_sqrt_price_x96(
                self.market_status.data.price,
                self.pool_info.token0.decimal,
                self.pool_info.token1.decimal,
                self.pool_info.is_token0_quote,
            )
        )
        delta_liquidity = (
            liquidity
            if (liquidity is not None) and liquidity < self.positions[position].liquidity
            else self.positions[position].liquidity
        )
        token0_get, token1_get = V3CoreLib.close_position(self._pool, position, delta_liquidity, sqrt_price_x96)

        self._positions[position].liquidity = self.positions[position].liquidity - delta_liquidity
        self._positions[position].pending_amount0 += token0_get
        self._positions[position].pending_amount1 += token1_get

        return token0_get, token1_get, delta_liquidity

    @write_func
    def __collect_fee(
        self,
        position: Position,
        max_collect_amount0: Decimal = None,
        max_collect_amount1: Decimal = None,
        collect_to_user: bool = True,
    ):
        """
        Collect fee

        :param position: get_position
        :param max_collect_amount0: max collect amount0
        :param max_collect_amount1: max collect amount1
        :param collect_to_user: Transfer collected token to user balance, default is true
        :return:
        """
        token0_fee = (
            max_collect_amount0
            if max_collect_amount0 is not None and max_collect_amount0 < position.pending_amount0
            else position.pending_amount0
        )
        token1_fee = (
            max_collect_amount1
            if max_collect_amount1 is not None and max_collect_amount1 < position.pending_amount1
            else position.pending_amount1
        )

        position.pending_amount0 -= token0_fee
        position.pending_amount1 -= token1_fee
        # add un_collect fee to current balance
        if collect_to_user:
            self.broker.add_to_balance(self.token0, token0_fee)
            self.broker.add_to_balance(self.token1, token1_fee)
        return token0_fee, token1_fee

    # action for strategy

    @float_param_formatter
    def add_liquidity(
        self,
        lower_quote_price: Decimal | float,
        upper_quote_price: Decimal | float,
        quote_max_amount: Decimal | float = None,
        base_max_amount: Decimal | float = None,
    ) -> (PositionInfo, Decimal, Decimal, int):
        """

        add liquidity, then get a new get_position

        :param lower_quote_price: lower price base on quote token.
        :type lower_quote_price: Decimal | float
        :param upper_quote_price: upper price base on quote token.
        :type upper_quote_price: Decimal | float
        :param base_max_amount:  inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type base_max_amount: Decimal | float
        :param quote_max_amount: inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type quote_max_amount: Decimal | float
        :return: added get_position, base token used, quote token used
        :rtype: (PositionInfo, Decimal, Decimal)
        """
        base_max_amount = self.broker.get_token_balance(self.base_token) if base_max_amount is None else base_max_amount
        quote_max_amount = (
            self.broker.get_token_balance(self.quote_token) if quote_max_amount is None else quote_max_amount
        )

        token0_amt, token1_amt = self._convert_pair(base_max_amount, quote_max_amount)
        lower_tick, upper_tick = V3CoreLib.quote_price_pair_to_tick(self._pool, lower_quote_price, upper_quote_price)
        lower_tick = nearest_usable_tick(lower_tick, self.pool_info.tick_spacing)
        upper_tick = nearest_usable_tick(upper_tick, self.pool_info.tick_spacing)
        # lower_tick, upper_tick = self._convert_pair(upper_tick, lower_tick)
        (
            created_position,
            token0_used,
            token1_used,
            liquidity,
        ) = self._add_liquidity_by_tick(token0_amt, token1_amt, lower_tick, upper_tick)
        base_used, quote_used = self._convert_pair(token0_used, token1_used)
        self._record_action(
            AddLiquidityAction(
                market=self.market_info,
                base_balance_after=self.broker.get_token_balance_with_unit(self.base_token),
                quote_balance_after=self.broker.get_token_balance_with_unit(self.quote_token),
                base_amount_max=UnitDecimal(base_max_amount, self.base_token.name),
                quote_amount_max=UnitDecimal(quote_max_amount, self.quote_token.name),
                lower_quote_price=UnitDecimal(lower_quote_price, self._pool_price_unit),
                upper_quote_price=UnitDecimal(upper_quote_price, self._pool_price_unit),
                base_amount_actual=UnitDecimal(base_used, self.base_token.name),
                quote_amount_actual=UnitDecimal(quote_used, self.quote_token.name),
                position=created_position,
                liquidity=int(liquidity),
            )
        )
        return created_position, base_used, quote_used, liquidity

    def add_liquidity_by_tick(
        self,
        lower_tick: int,
        upper_tick: int,
        base_max_amount: Decimal | float = None,
        quote_max_amount: Decimal | float = None,
        sqrt_price_x96: int = -1,
        tick: int = -1,
        trim_tick: bool = True,
    ) -> (PositionInfo, Decimal, Decimal, int):
        """

        add liquidity, you need to set tick instead of price.

        :param lower_tick: lower tick
        :type lower_tick: int
        :param upper_tick: upper tick
        :type upper_tick: int
        :param base_max_amount:  inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type base_max_amount: Decimal | float
        :param quote_max_amount: inputted base token amount, also the max amount to deposit, if is None, will use all the balance of base token
        :type quote_max_amount: Decimal | float
        :param tick: tick price.  if set to none, it will be calculated from current price.
        :type tick: int
        :param sqrt_price_x96: precise price.  if set to none, it will be calculated from current price. this param will override tick
        :type sqrt_price_x96: int
        :param trim_tick: trim tick according to tick spacing, default is True
        :type trim_tick: bool
        :return: added get_position, base token used, quote token used, liquidity
        :rtype: (PositionInfo, Decimal, Decimal, int)
        """
        if trim_tick:
            lower_tick = nearest_usable_tick(lower_tick, self.pool_info.tick_spacing)
            upper_tick = nearest_usable_tick(upper_tick, self.pool_info.tick_spacing)

        if lower_tick > upper_tick:
            lower_tick, upper_tick = upper_tick, lower_tick

        if sqrt_price_x96 == -1 and tick != -1:
            sqrt_price_x96 = tick_to_sqrt_price_x96(tick)

        base_max_amount = self.broker.get_token_balance(self.base_token) if base_max_amount is None else base_max_amount
        quote_max_amount = (
            self.broker.get_token_balance(self.quote_token) if quote_max_amount is None else quote_max_amount
        )

        token0_amt, token1_amt = self._convert_pair(base_max_amount, quote_max_amount)
        (
            created_position,
            token0_used,
            token1_used,
            liquidity,
        ) = self._add_liquidity_by_tick(token0_amt, token1_amt, lower_tick, upper_tick, sqrt_price_x96)
        base_used, quote_used = self._convert_pair(token0_used, token1_used)
        self._record_action(
            AddLiquidityAction(
                market=self.market_info,
                base_balance_after=self.broker.get_token_balance_with_unit(self.base_token),
                quote_balance_after=self.broker.get_token_balance_with_unit(self.quote_token),
                base_amount_max=UnitDecimal(base_max_amount, self.base_token.name),
                quote_amount_max=UnitDecimal(quote_max_amount, self.quote_token.name),
                lower_quote_price=UnitDecimal(self.tick_to_price(lower_tick), self._pool_price_unit),
                upper_quote_price=UnitDecimal(self.tick_to_price(upper_tick), self._pool_price_unit),
                base_amount_actual=UnitDecimal(base_used, self.base_token.name),
                quote_amount_actual=UnitDecimal(quote_used, self.quote_token.name),
                position=created_position,
                liquidity=int(liquidity),
            )
        )
        return created_position, base_used, quote_used, liquidity

    @float_param_formatter
    def remove_liquidity(
        self,
        position: PositionInfo,
        liquidity: int = None,
        collect: bool = True,
        sqrt_price_x96: int = -1,
        remove_dry_pool: bool = True,
    ) -> (Decimal, Decimal):
        """
        | remove liquidity from the pool. Liquidity will be reduced to 0,
        | instead of send tokens to broker, tokens will be transferred to fee property in get_position.
        | get_position will be not deleted, until fees and tokens are collected.

        :param position: get_position to remove.
        :type position: PositionInfo
        :param liquidity: liquidity amount to remove, if set to None, all the liquidity will be removed
        :type liquidity: int
        :param collect: collect or not, if collect, will call collect function. and tokens will be sent to broker. if not, token will be kept in fee property of postion
        :type collect: bool
        :param sqrt_price_x96: precise price.  if set to none, it will be calculated from current price.
        :type sqrt_price_x96: int
        :param remove_dry_pool: remove pool which liquidity==0, effect when collect==True
        :type remove_dry_pool: bool
        :return: (base_got,quote_get), base and quote token amounts collected from get_position
        :rtype:  (Decimal,Decimal)
        """
        if liquidity and liquidity < 0:
            raise DemeterError("liquidity should large than 0")
        token0_get, token1_get, delta_liquidity = self.__remove_liquidity(position, liquidity, sqrt_price_x96)

        base_get, quote_get = self._convert_pair(token0_get, token1_get)
        self._record_action(
            RemoveLiquidityAction(
                market=self.market_info,
                base_balance_after=self.broker.get_token_balance_with_unit(self.base_token),
                quote_balance_after=self.broker.get_token_balance_with_unit(self.quote_token),
                position=position,
                base_amount=UnitDecimal(base_get, self.base_token.name),
                quote_amount=UnitDecimal(quote_get, self.quote_token.name),
                removed_liquidity=delta_liquidity,
                remain_liquidity=self.positions[position].liquidity,
            )
        )
        if collect:
            return self.collect_fee(position, remove_dry_pool=remove_dry_pool)
        else:
            return base_get, quote_get

    @float_param_formatter
    def collect_fee(
        self,
        position: PositionInfo,
        max_collect_amount0: Decimal = None,
        max_collect_amount1: Decimal = None,
        remove_dry_pool: bool = True,
        collect_to_user: bool = True,
    ) -> (Decimal, Decimal):
        """
        | collect fee and token from positions,
        | if the amount and liquidity is zero, this get_position will be deleted.

        :param position: get_position to collect
        :type position: PositionInfo
        :param max_collect_amount0: max token0 amount to collect, e.g. 1.2345 usdc, if set to None, all the amount will be collect
        :type max_collect_amount0: Decimal
        :param max_collect_amount1: max token0 amount to collect, if set to None, all the amount will be collect
        :type max_collect_amount1: Decimal
        :param remove_dry_pool: remove pool which liquidity==0, effect when collect==True
        :type remove_dry_pool: bool
        :param collect_to_user: Transfer collected token to user balance, default is true
        :type collect_to_user: bool
        :return: (base_got,quote_get), base and quote token amounts collected from get_position
        :rtype:  (Decimal,Decimal)
        """
        if (max_collect_amount0 and max_collect_amount0 < 0) or (max_collect_amount1 and max_collect_amount1 < 0):
            raise DemeterError("collect amount should large than 0")
        token0_get, token1_get = self.__collect_fee(
            self._positions[position], max_collect_amount0, max_collect_amount1, collect_to_user
        )

        base_get, quote_get = self._convert_pair(token0_get, token1_get)
        if self._positions[position]:
            self._record_action(
                CollectFeeAction(
                    market=self.market_info,
                    base_balance_after=self.broker.get_token_balance_with_unit(self.base_token),
                    quote_balance_after=self.broker.get_token_balance_with_unit(self.quote_token),
                    position=position,
                    base_amount=UnitDecimal(base_get, self.base_token.name),
                    quote_amount=UnitDecimal(quote_get, self.quote_token.name),
                )
            )
        if (
            self._positions[position].pending_amount0 == Decimal(0)
            and self._positions[position].pending_amount1 == Decimal(0)
            and self._positions[position].liquidity == 0
            and remove_dry_pool
        ):
            del self.positions[position]
        return base_get, quote_get

    @float_param_formatter
    def swap(
        self,
        from_amount: Decimal | float,
        from_token: TokenInfo,
        to_token: TokenInfo,
        price: Decimal | float = None,
        throw_action=True,
    ):
        if from_token == to_token:
            raise DemeterError("from and to token can not same")
        if from_token not in [self.quote_token, self.base_token] or to_token not in [self.quote_token, self.base_token]:
            raise DemeterError("from or to token not in pool")

        if from_token == self.base_token:
            # e.g. swap 1 eth for 3000 usdc
            price = price if price else self.market_status.data.price
        else:
            # e.g. swap 3000 usdc for 1 eth
            price = price if price else 1 / self.market_status.data.price
        fee_in_from = from_amount * self.pool_info.fee_rate
        to_amount = (from_amount - fee_in_from) * price
        self.broker.subtract_from_balance(from_token, from_amount)
        self.broker.add_to_balance(to_token, to_amount)
        if throw_action:
            self._record_action(
                SwapAction(
                    market=self.market_info,
                    amount=UnitDecimal(from_amount, from_token.name),
                    price=UnitDecimal(price, f"{from_token.name}/{to_token.name}"),
                    fee=UnitDecimal(fee_in_from, from_token.name),
                    to_amount=UnitDecimal(to_amount, to_token.name),
                )
            )
        return fee_in_from, to_amount

    @float_param_formatter
    def buy(self, base_token_amount: Decimal | float, price: Decimal | float = None) -> (Decimal, Decimal, Decimal):
        """
        buy base token, swap from quote token to base token.

        :param base_token_amount: amount to buy(in base token)
        :type base_token_amount:  Decimal | float
        :param price: price in base/quote, e.g. 1234 eth/usdc, if leave to None, will use pool price
        :type price: Decimal | float
        :return: fee in base token, quote token amount spend, base token amount got
        :rtype: (Decimal, Decimal, Decimal)
        """
        if base_token_amount == 0:
            return DECIMAL_0, DECIMAL_0, DECIMAL_0
        price = price if price else self.market_status.data.price
        quote_amount_with_fee = base_token_amount * price / (1 - self._pool.fee_rate)
        fee_in_quote, base_amount_got = self.swap(
            quote_amount_with_fee, self.quote_token, self.base_token, 1 / price, False
        )
        self._record_action(
            BuyAction(
                market=self.market_info,
                base_balance_after=self.broker.get_token_balance_with_unit(self.base_token),
                quote_balance_after=self.broker.get_token_balance_with_unit(self.quote_token),
                amount=UnitDecimal(base_token_amount, self.base_token.name),
                price=UnitDecimal(price, self._pool_price_unit),
                fee=UnitDecimal(fee_in_quote, self.quote_token.name),
                base_change=UnitDecimal(base_amount_got, self.base_token.name),
                quote_change=UnitDecimal(quote_amount_with_fee, self.quote_token.name),
            )
        )
        return fee_in_quote, quote_amount_with_fee, base_amount_got

    @float_param_formatter
    def sell(self, base_token_amount: Decimal | float, price: Decimal | float = None) -> (Decimal, Decimal, Decimal):
        """
        Sell base token, swap from base token to quote token.

        :param base_token_amount: amount to sell(in base token)
        :type base_token_amount:  Decimal | float
        :param price: price, e.g. 1234 eth/usdc, if leave to None, will use pool price
        :type price: Decimal | float
        :return: fee in quote token, base token amount spend, quote token amount got
        :rtype: (Decimal, Decimal, Decimal)
        """
        if base_token_amount == 0:
            return DECIMAL_0, DECIMAL_0, DECIMAL_0
        price = price if price else self.market_status.data.price

        fee_in_base, quote_amount_got = self.swap(base_token_amount, self.base_token, self.quote_token, price, False)

        self._record_action(
            SellAction(
                market=self.market_info,
                base_balance_after=self.broker.get_token_balance_with_unit(self.base_token),
                quote_balance_after=self.broker.get_token_balance_with_unit(self.quote_token),
                amount=UnitDecimal(base_token_amount, self.base_token.name),
                price=UnitDecimal(price, self._pool_price_unit),
                fee=UnitDecimal(fee_in_base, self.base_token.name),
                base_change=UnitDecimal(base_token_amount, self.base_token.name),
                quote_change=UnitDecimal(quote_amount_got, self.quote_token.name),
            )
        )

        return fee_in_base, base_token_amount, quote_amount_got

    def add_liquidity_by_value(
        self, lower_tick: int, upper_tick: int, value_to_use: Decimal | None = None, trim_tick: bool = True
    ) -> (PositionInfo, Decimal, Decimal, int):
        """
        Add liquidity from balance with value defined, and swap if necessary.
        e.g. you have 1 eth and 3000 usdc, and eth price is 1000. If you want to invest at 1:1,
        you have to swap 1000 usdc to eth, so your balance will be 2 eth and 2000 usdc, but swap fee is not counted.
        this function will calculate the best swap amount(1005.025, if fee is 1%),
        which will make the balance at: 1.99975 eth and 1994.975 usdc, which is just 1:1, So all balance will be used.
        This function to free you from complex calculation.


        :param lower_tick: lower tick
        :type lower_tick: int
        :param upper_tick: upper tick
        :type upper_tick: int
        :param trim_tick: trim tick according to tick spacing, default is True
        :type trim_tick: bool
        :param value_to_use: Value you want to add liquidity(in quote token). Actual value used would be less than this value because swap fee might be charged.
        :return: added get_position, base token used, quote token used, liquidity
        :rtype: (PositionInfo, Decimal, Decimal, int)
        """
        if trim_tick:
            lower_tick = nearest_usable_tick(lower_tick, self.pool_info.tick_spacing)
            upper_tick = nearest_usable_tick(upper_tick, self.pool_info.tick_spacing)
        price = self._market_status.data.price
        tick = self._market_status.data.closeTick
        price0, price1 = self._convert_pair(price, Decimal(1))

        quote_balance = self.broker.get_token_balance(self.quote_token)
        base_balance_value = self.broker.get_token_balance(self.base_token) * price

        balance = quote_balance + base_balance_value

        if value_to_use is None:
            value_to_use = balance

        if value_to_use > balance:
            raise DemeterError("Not enough balance to add liquidity")
        if lower_tick >= upper_tick:
            raise DemeterError("Lower tick is larger than upper tick")

        # price is greater than upper price
        if (self._is_token0_quote and tick > upper_tick) or (not self._is_token0_quote and tick < lower_tick):
            # all base
            base_amount = value_to_use / self._market_status.data.price
            diff = base_amount - self.broker.get_token_balance(self.base_token)
            fee_in_quote = Decimal(0)
            if diff > MIN_ERROR:
                fee_in_quote, to_amount = self.swap(diff * price, self.quote_token, self.base_token)
            return self.add_liquidity_by_tick(lower_tick, upper_tick, base_amount - fee_in_quote / price, Decimal(0))

        # price is lower than upper price
        if (self._is_token0_quote and tick < lower_tick) or (not self._is_token0_quote and tick > upper_tick):
            # all quote
            diff = value_to_use - self.broker.get_token_balance(self.quote_token)
            fee_in_base = Decimal(0)
            if diff > 0:
                fee_in_base, quote_got = self.swap(diff / price, self.base_token, self.quote_token)
            return self.add_liquidity_by_tick(lower_tick, upper_tick, Decimal(0), value_to_use - fee_in_base * price)

        # price is in tick range

        ratio = estimate_ratio(tick, lower_tick, upper_tick)
        ratio_in_amount = ratio * 10 ** (self.pool_info.token1.decimal - self.pool_info.token0.decimal)
        ratio_in_value = Decimal(ratio_in_amount) / price if self._is_token0_quote else Decimal(ratio_in_amount) * price

        token1_value = value_to_use / (ratio_in_value + 1)
        token0_value = value_to_use - token1_value
        balance0_value = self.broker.get_token_balance(self.token0) * price0
        balance1_value = self.broker.get_token_balance(self.token1) * price1
        if token0_value <= balance0_value and token1_value <= balance1_value:
            # do not need swap
            base_value, quote_value = self._convert_pair(token0_value, token1_value)
            return self.add_liquidity_by_tick(lower_tick, upper_tick, base_value / price, quote_value)
        elif token0_value > balance0_value and token1_value > balance1_value:
            raise DemeterError("Not enough balance to add liquidity")
        elif token0_value < balance0_value and token1_value > balance1_value:
            # need swap from token0 to token1
            actual_token0_value, actual_token1_value, swap_value = get_swap_value_with_part_balance_used(
                balance0_value, balance1_value, value_to_use, self.pool_info.fee_rate, ratio_in_value
            )
            if self._is_token0_quote:
                self.swap(swap_value, self.quote_token, self.base_token)
            else:
                self.swap(swap_value / price, self.base_token, self.quote_token)
            base_value, quote_value = self._convert_pair(actual_token0_value, actual_token1_value)
            return self.add_liquidity_by_tick(lower_tick, upper_tick, base_value / price, quote_value)
        elif token0_value > balance0_value and token1_value < balance1_value:
            # need to swap from token1 to token 0
            actual_token1_value, actual_token0_value, swap_value = get_swap_value_with_part_balance_used(
                balance1_value, balance0_value, value_to_use, self.pool_info.fee_rate, 1 / ratio_in_value
            )
            if self._is_token0_quote:
                self.swap(swap_value / price, self.base_token, self.quote_token)
            else:
                self.swap(swap_value, self.quote_token, self.base_token)
            base_value, quote_value = self._convert_pair(actual_token0_value, actual_token1_value)
            return self.add_liquidity_by_tick(lower_tick, upper_tick, base_value / price, quote_value)
        else:
            raise NotImplementedError()
        pass

    def even_rebalance(self, price: Decimal | None = None):
        """
        Divide assets equally between two tokens.

        :param price: price of quote token. e.g. 1234 eth/usdc, if leave to None, will use pool price
        :type price: Decimal
        """
        if price is None:
            price = self._market_status.data.price

        amount_quote = self.broker.get_token_balance(self.quote_token)
        amount_base = self.broker.get_token_balance(self.base_token)

        delta_base = (amount_quote / price - amount_base) / (Decimal(2) + self.pool_info.fee_rate)
        if delta_base >= 0:
            self.buy(delta_base)
            return

        delta_quote = (amount_base - amount_quote / price) / (Decimal(2) - self.pool_info.fee_rate)
        if delta_quote >= 0:
            self.sell(delta_quote)
            return
        pass

    def remove_all_liquidity(self):
        """
        remove all the positions kept in broker.
        """
        if len(self.positions) < 1:
            return
        keys = list(self.positions.keys())
        for position_key in keys:
            self.remove_liquidity(position_key)

    def add_statistic_column(self, df: pd.DataFrame):
        """
        add statistic column to data, new columns including:

        * open: open price
        * price: close price (current price)
        * low: lowest price
        * high: height price
        * volume0: swap volume for token 0
        * volume1: swap volume for token 1

        :param df: original data
        :type df: pd.DataFrame

        """
        # add statistic column
        df["open"] = df["openTick"].map(lambda x: self.tick_to_price(x))
        df["price"] = df["closeTick"].map(lambda x: self.tick_to_price(x))
        high_name, low_name = (
            ("lowestTick", "highestTick") if self.pool_info.is_token0_quote else ("highestTick", "lowestTick")
        )
        df["low"] = df[high_name].map(lambda x: self.tick_to_price(x))
        df["high"] = df[low_name].map(lambda x: self.tick_to_price(x))
        df["volume0"] = df["inAmount0"].map(lambda x: Decimal(x) / 10**self.pool_info.token0.decimal)
        df["volume1"] = df["inAmount1"].map(lambda x: Decimal(x) / 10**self.pool_info.token1.decimal)

    def load_data(self, chain: str, contract_addr: str, start_date: date, end_date: date):
        """

        load data, and preprocess. preprocess actions including:

        * fill empty data
        * calculate statistic column
        * set timestamp as index

        :param chain: chain name
        :type chain: str
        :param contract_addr: pool contract address
        :type contract_addr: str
        :param start_date: start test date
        :type start_date: date
        :param end_date: end test date
        :type end_date: date
        """
        cache_key = CacheManager.get_cache_key(self.market_info.type.name, start_date, end_date, chain, contract_addr)
        cache_df = CacheManager.load(cache_key)
        if cache_df is not None:
            self.data = cache_df
            return

        self.logger.info(f"start load files from {start_date} to {end_date}...")
        df = pd.DataFrame()
        day = start_date
        if start_date > end_date:
            raise DemeterError(f"start date {start_date} should earlier than end date {end_date}")
        while day <= end_date:
            new_type_path = os.path.join(
                self.data_path,
                f"{chain.lower()}-{contract_addr}-{day.strftime('%Y-%m-%d')}.minute.csv",
            )
            path = (
                new_type_path
                if os.path.exists(new_type_path)
                else os.path.join(
                    self.data_path,
                    f"{chain}-{contract_addr}-{day.strftime('%Y-%m-%d')}.csv",
                )
            )
            if not os.path.exists(path):
                raise IOError(
                    f"resource file {new_type_path} not found, please download with demeter-fetch: https://github.com/zelos-alpha/demeter-fetch"
                )
            day_df = pd.read_csv(
                path,
                converters={
                    "inAmount0": to_decimal,
                    "inAmount1": to_decimal,
                    "netAmount0": to_decimal,
                    "netAmount1": to_decimal,
                    "currentLiquidity": to_decimal,
                },
            )
            if len(day_df.index) > 0:
                df = pd.concat([df, day_df])
            day = day + timedelta(days=1)
        self.logger.info("load file complete, preparing...")

        df["timestamp"] = pd.to_datetime(df["timestamp"])
        df.set_index("timestamp", inplace=True)

        # fill empty row (first minutes in a day, might be blank)
        full_indexes = pd.date_range(
            start=start_date,
            end=datetime.combine(end_date, time(0, 0, 0)) + timedelta(days=1) - timedelta(minutes=1),
            freq="1min",
        )
        df = df.reindex(full_indexes)
        # df = Lines.from_dataframe(df)
        # df = df.fillna()
        df: pd.DataFrame = fillna(df)
        if pd.isna(df.iloc[0]["closeTick"]):
            df = df.bfill()

        self.add_statistic_column(df)
        self.data = df
        CacheManager.save(cache_key, df)
        self.logger.info("data has been prepared")

    def formatted_str(self) -> str:
        """
        Return a brief description of this market in pretty format. Used for print in console.
        """
        value = get_formatted_predefined(f"{self.market_info.name}({type(self).__name__})", STYLE["header3"]) + "\n"
        value += (
            get_formatted_from_dict(
                {
                    "token0": self.pool_info.token0.name,
                    "token1": self.pool_info.token1.name,
                    "fee(%)": self.pool_info.fee_rate * 100,
                    "quote token": self.quote_token.name,
                }
            )
            + "\n"
        )
        value += get_formatted_predefined("positions", STYLE["key"]) + "\n"
        df = position_dict_to_dataframe(self.positions)
        if len(df.index) > 0:
            value += df.to_string()
        else:
            value += "Empty DataFrame\n"
        return value

    def _resample(self, freq: str):
        self._data = resample(self.data, freq)



from ._typing import (
    Vault,
    SqueethBalance,
    VaultKey,
    AddVaultAction,
    UpdateCollateralAction,
    UpdateShortAction,
    DepositLpAction,
    WithdrawLpAction,
    ReduceDebtAction,
    LiquidationAction,
    SqueethDescription,
)
from .market import SqueethMarket
from .helper import calc_twap_price



import json
import pandas as pd
from decimal import Decimal
from enum import Enum
from functools import wraps
from types import SimpleNamespace
from typing import Any, Dict

from demeter import TokenInfo, STABLE_COINS
from demeter._typing import USD

OUTPUT_WIDTH = 30


def orjson_default(obj):
    if isinstance(obj, Decimal):
        return str(obj)
    raise TypeError


def to_decimal(value: Any) -> Decimal:
    """
    convert value to decimal

    :param value: any value
    :type value: Any
    :return: Decimal value
    :rtype: Decimal

    """
    return Decimal(str(value))


def object_to_decimal(num: Any) -> Any:
    """
    If number is float or int, return Decimal, else return original value

    :param value: any value
    :type value: Any
    :return: Decimal value
    :rtype: Any
    """
    return Decimal(str(num)) if (isinstance(num, float) or type(num) == int) else num


def dict_to_object(dict_entity: Dict) -> Any:
    """
    convert dict to object via json

    :param dict_entity: dict instance
    :type dict_entity: Dict
    :return: object
    :rtype: Any
    """
    return json.loads(json.dumps(dict_entity), object_hook=lambda d: SimpleNamespace(**d))


def float_param_formatter(func):
    """
    decorator to convert param to float

    :param func: any function
    :return: function execute result
    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        new_args = ()
        for arg in args:
            new_args += (object_to_decimal(arg),)
        for k, v in kwargs.items():
            kwargs[k] = object_to_decimal(v)
        return func(*new_args, **kwargs)

    return wrapper_func


def get_enum_by_name(me: Enum, name: str):
    """
    get enum by name

    :param me: enum class
    :param name: enum item name
    :return: get value in enum
    """
    for e in me:
        if e.name.lower() == name.lower():
            return e
    raise RuntimeError(f"cannot found {name} in {me}, allow value is " + str([x.name for x in me]))


def require(condition: bool, error_msg: str):
    """
    Checking whether the condition is True, if not, will raise a AssertionError

    :param condition: condition
    :type condition: bool
    :param error_msg: error message contains in AssertionError
    :param error_msg: str
    """
    if not condition:
        raise AssertionError(error_msg)


def to_multi_index_df(df: pd.DataFrame, level0: str):
    new_level_0 = [level0] * df.shape[1]
    new_columns = pd.MultiIndex.from_arrays([new_level_0, df.columns], names=["l1", "l2"])
    df.columns = new_columns


def load_account_status(path: str) -> pd.DataFrame:
    if path.endswith(".csv"):
        df = pd.read_csv(path, index_col=[0], header=[0, 1], parse_dates=[0])
        rename_dict = {}
        for column in df.columns:
            if "Unnamed" in column[1]:
                rename_dict[column[1]] = ""
        df = df.rename(columns=rename_dict, level=1)
    elif path.endswith(".pkl"):
        df = pd.read_pickle(path, compression="gzip")
    else:
        raise RuntimeError("Unknown format, account status support csv and pickle only")
    return df


def is_stable_coin(*token: TokenInfo):
    """
    check token list has stable token
    :param token:
    :return: the first stable token, or none
    """
    for t in token:
        if t == USD or t.name in STABLE_COINS:
            return t
    return None



"""
Broker supports different order types, and track cash and positions.
"""

from ._typing import (
    BaseAction,
    MarketBalance,
    AccountStatus,
    MarketInfo,
    AccountStatusCommon,
    Asset,
    MarketDict,
    AssetDict,
    ActionTypeEnum,
    MarketStatus,
    Rule,
    MarketTypeEnum,
    RowData,
    BASE_FREQ,
)
from .broker import Broker
from .market import Market, write_func



import pandas as pd
from _decimal import Decimal
from dataclasses import dataclass
from enum import Enum
from typing import Dict, NamedTuple, Union
from typing import TypeVar

from .. import TokenInfo, UnitDecimal
from .._typing import MarketDescription
from ..broker import MarketBalance, MarketStatus, BaseAction, ActionTypeEnum
from ..utils import console_text
from ..utils.console_text import get_action_str, ForColorEnum

T = TypeVar("T")
K = TypeVar("K")


class InterestRateMode(Enum):
    """
    Interest rate mode
    """

    variable = 1
    stable = 2

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


@dataclass
class ActionKey:
    """
    Abstract key for actions(supply, borrow, etc.)

    :param token: token of this position
    :type token: TokenInfo
    """

    token: TokenInfo
    """token of this position"""

    def __str__(self):
        return self.token.name

    def __hash__(self):
        return self.token.__hash__()


@dataclass
class BorrowKey(ActionKey):
    """
    key of dict for borrow actions

    :param token: token of this position
    :type token: TokenInfo
    :param interest_rate_mode: interest rate mode
    :type interest_rate_mode: InterestRateMode
    """

    interest_rate_mode: InterestRateMode
    """interest rate mode"""

    def __str__(self):
        return f"Borrow_{self.token.name}({self.interest_rate_mode.name})"

    def __repr__(self):
        return f"Borrow_{self.token.name}({self.interest_rate_mode.name})"

    def __hash__(self):
        return hash((self.token, self.interest_rate_mode))


@dataclass
class SupplyKey(ActionKey):
    """
    key of dict for supply actions

    :param token: token of this position
    :type token: TokenInfo
    """

    def __str__(self):
        return "Supply_" + self.token.name

    def __repr__(self):
        return "Supply_" + self.token.name

    def __hash__(self):
        return self.token.__hash__()


@dataclass
class SupplyInfo:
    """
    Basic info for supply, designed to kept properties inside market

    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index
    :type base_amount: Decimal
    :param collateral: set this supply to collateral or not
    :type collateral: bool

    """

    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index"""
    collateral: bool
    """set this supply to collateral or not"""
    begin_supply_index: Decimal
    """index value when do supply"""


@dataclass
class Supply:
    """
    Supply info, designed to show supplies to user

    :param token: which token is supplied
    :type token: TokenInfo
    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index_at_supply_moment
    :type base_amount: Decimal
    :param collateral: set this supply to collateral or not
    :type collateral: bool
    :param amount: Actual amount at this moment. value is amount * liquidity_index_at_this_moment, unit in token amount, e.g. 1 eth
    :type amount: Decimal
    :param apy: current apy (annual interest rate)
    :type apy: Decimal
    :param value: value at this moment. unit is usd, e.g. 1400 usd
    :type value: Decimal
    """

    token: TokenInfo
    """which token is supplied"""
    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index_at_supply_moment"""
    collateral: bool
    """set this supply to collateral or not"""
    amount: Decimal
    """Actual amount at this moment. value is amount * liquidity_index_at_this_moment, unit in token amount, e.g. 1 eth"""
    apy: Decimal
    """current apy (annual interest rate)"""
    value: Decimal
    """value at this moment. unit is usd, e.g. 1400 usd"""
    begin_supply_index: Decimal
    """index value when do supply"""

def supply_to_dataframe(supplies: Dict[SupplyKey, Supply]) -> pd.DataFrame:
    """
    convert supply dict to a dataframe
    """
    pos_dict = {
        "token": [],
        "base_amount": [],
        "collateral": [],
        "amount": [],
        "apy": [],
        "value": [],
    }
    for k, v in supplies.items():
        pos_dict["token"].append(v.token.name)
        pos_dict["base_amount"].append(console_text.format_value(v.base_amount))
        pos_dict["collateral"].append(v.collateral)
        pos_dict["amount"].append(console_text.format_value(v.amount))
        pos_dict["apy"].append(console_text.format_value(v.apy))
        pos_dict["value"].append((console_text.format_value(v.value)))
    return pd.DataFrame(pos_dict)


@dataclass
class BorrowInfo:
    """
    Basic info for borrow, designed to kept properties inside market

    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index
    :type base_amount: Decimal
    """

    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/liquidity_index"""
    begin_borrow_index: Decimal
    """borrow index value when borrow"""


@dataclass
class Borrow:
    """
    borrow info, designed to show borrows to user

    :param token: which token is borrowed
    :type token: TokenInfo
    :param base_amount: Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/variable_borrow_index_at_supply_moment
    :type base_amount: Decimal
    :param interest_rate_mode: Interest rate mode
    :type interest_rate_mode: InterestRateMode
    :param amount: Actual amount at this moment. value is amount * variable_borrow_index_at_this_moment, unit in token amount, e.g. 1 eth
    :type amount: Decimal
    :param apy:  current apy (annual interest rate)
    :type apy: Decimal
    :param value: value at this moment. unit is usd, e.g. 1400 usd
    :type value: Decimal
    """

    token: TokenInfo
    """which token is borrowed"""
    base_amount: Decimal
    """Base amount of token. Note: base amount is the amount kept in aave contract. its value is amount/variable_borrow_index_at_supply_moment"""
    interest_rate_mode: InterestRateMode
    """Interest rate mode"""
    amount: Decimal
    """Actual amount at this moment. value is amount * variable_borrow_index_at_this_moment, unit in token amount, e.g. 1 eth"""
    apy: Decimal
    """current apy (annual interest rate)"""
    value: Decimal
    """value at this moment. unit is usd, e.g. 1400 usd"""
    begin_borrow_index: Decimal
    """borrow index value when borrow"""

def borrow_to_dataframe(supplies: Dict[BorrowKey, Borrow]) -> pd.DataFrame:
    """
    convert borrow dict to a dataframe

    """
    pos_dict = {
        "token": [],
        "base_amount": [],
        "mode": [],
        "amount": [],
        "apy": [],
        "value": [],
    }
    for k, v in supplies.items():
        pos_dict["token"].append(v.token.name)
        pos_dict["base_amount"].append(console_text.format_value(v.base_amount))
        pos_dict["mode"].append(v.interest_rate_mode.name)
        pos_dict["amount"].append(console_text.format_value(v.amount))
        pos_dict["apy"].append(console_text.format_value(v.apy))
        pos_dict["value"].append(console_text.format_value(v.value))
    return pd.DataFrame(pos_dict)


@dataclass
class AaveBalance(MarketBalance):
    """
    Asset or position value in aave market.

    :param supplies_count: count of supplies
    :type supplies_count: int
    :param borrows_count: count of borrows
    :type borrows_count: int
    :param borrows_value: total borrow value(in usd)
    :type borrows_value: Decimal
    :param supplies_value: total supply value(in usd)
    :type supplies_value: Decimal
    :param collaterals_value: total collateral value in supplies(in usd)
    :type collaterals_value: Decimal
    :param health_factor: current health factor
    :type health_factor: Decimal
    :param current_ltv: max ltv allowed, in decimal, e.g. 0.7568
    :type current_ltv: Decimal
    :param liquidation_threshold: current liquidation threshold, in decimal, e.g.0.8
    :type liquidation_threshold: Decimal
    :param supply_apy: annual interest rate of supplies
    :type supply_apy: Decimal
    :param borrow_apy: annual interest rate of borrows
    :type borrow_apy: Decimal
    :param net_apy: total annual interest rate of all supplies/borrows
    :type net_apy: Decimal
    """

    supplies_count: int
    """count of supplies"""
    borrows_count: int
    """count of borrows"""
    borrows_value: Decimal
    """total borrow value(in usd)"""
    supplies_value: Decimal
    """total supply value(in usd)"""
    collaterals_value: Decimal
    """total collateral value in supplies(in usd)"""
    health_factor: Decimal
    """current health factor"""
    current_ltv: Decimal
    """max ltv allowed, in decimal, e.g. 0.7568"""
    liquidation_threshold: Decimal
    """current liquidation threshold, in decimal, e.g.0.8"""
    supply_apy: Decimal
    """annual interest rate of supplies"""
    borrow_apy: Decimal
    """annual interest rate of borrows"""
    net_apy: Decimal
    """total annual interest rate of all supplies/borrows"""


@dataclass
class AaveTokenStatus:
    """
    Aave pool status of one token. Usually they are got form ReserveDataUpdated event of aave pool. Consider it's the transient interest rates

    :param liquidity_rate: interest rate of supply in a second
    :type liquidity_rate: Decimal
    :param stable_borrow_rate: interest rate of stable borrow in a second
    :type stable_borrow_rate: Decimal
    :param variable_borrow_rate: interest rate of variable borrow rate in a second
    :type variable_borrow_rate: Decimal
    :param liquidity_index: Decide supply amount at this moment. consider it's the average interest rate
    :type liquidity_index: Decimal
    :param variable_borrow_index: Decide borrow amount at this moment. consider it's the average borrow interest rate
    :type variable_borrow_index: Decimal
    """

    liquidity_rate: Decimal
    """interest rate of supply in a second"""
    stable_borrow_rate: Decimal
    """interest rate of stable borrow in a second"""
    variable_borrow_rate: Decimal
    """interest rate of variable borrow rate in a second"""
    liquidity_index: Decimal
    """Decide supply amount at this moment. consider it's the average interest rate"""
    variable_borrow_index: Decimal
    """Decide borrow amount at this moment. consider it's the average borrow interest rate"""


@dataclass
class AaveMarketStatus(MarketStatus):
    """
    MarketStatus properties

    :param data: pool status
    :type data: Union[pd.Series, AaveTokenStatus]
    """

    data: Union[pd.Series, AaveTokenStatus] = None
    """pool status"""


class RiskParameter:
    """
    | Risk parameter of tokens,
    | It's a column description of dataframe
    | The csv is downloaded from https://www.config.fyi/,
    | Note: some attributes in csv file are excluded.
    """

    symbol: str
    canCollateral: bool
    LTV: float
    liqThereshold: float
    liqBonus: float
    reserveFactor: float
    canBorrow: bool
    optimalUtilization: float
    canBorrowStable: bool
    debtCeiling: float
    supplyCap: float
    borrowCap: float
    eModeLtv: float
    eModeLiquidationThereshold: float
    eModeLiquidationBonus: float
    borrowableInIsolation: float


@dataclass
class SupplyAction(BaseAction):
    """
    Describe parameters and results of supply transaction

    :param token: which token is supplied
    :type token: str
    :param amount: amount supplied
    :type amount: UnitDecimal
    :param collateral: collateral the supply or not.
    :type collateral: bool
    :param deposit_after: total supply amount of this token after supply
    :type deposit_after: UnitDecimal
    """

    token: str
    """which token is supplied"""
    amount: UnitDecimal
    """amount supplied"""
    collateral: bool
    """collateral the supply or not."""
    deposit_after: UnitDecimal
    """total supply amount of this token after supply"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_supply

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.light_green,
            {
                "token": self.token,
                "amount": self.amount.to_str(),
                "collateral": str(self.collateral),
                "deposit_after": self.deposit_after.to_str(),
            },
        )


@dataclass
class WithdrawAction(BaseAction):
    """
    Describe parameters and results of withdraw transaction

    :param token: which token is supplied
    :type token: str
    :param amount: amount supplied
    :type amount: UnitDecimal
    :param deposit_after: total supply amount of this token after withdraw
    :type deposit_after: UnitDecimal
    """

    token: str
    """which token is supplied"""
    amount: UnitDecimal
    """amount supplied"""
    deposit_after: UnitDecimal
    """total supply amount of this token after withdraw"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_withdraw

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.light_red,
            {
                "token": self.token,
                "amount": self.amount.to_str(),
                "deposit_after": self.deposit_after.to_str(),
            },
        )


@dataclass
class BorrowAction(BaseAction):
    """
    Describe parameters and results of borrow transaction

    :param token: which token is borrowed
    :type token: str
    :param interest_rate_mode: interest rate mode
    :type interest_rate_mode: InterestRateMode
    :param amount: amount borrowed
    :type amount: UnitDecimal
    :param debt_after: total borrow amount of this token after borrow transaction
    :type debt_after: UnitDecimal
    """

    token: str
    """which token is borrowed"""
    interest_rate_mode: InterestRateMode
    """interest rate mode"""
    amount: UnitDecimal
    """amount borrowed"""
    debt_after: UnitDecimal
    """total borrow amount of this token after borrow transaction"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_borrow

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.green,
            {
                "token": self.token,
                "interest_rate_mode": self.interest_rate_mode.name,
                "amount": self.amount.to_str(),
                "debt_after": self.debt_after.to_str(),
            },
        )


@dataclass
class RepayAction(BaseAction):
    """
    Describe parameters and results of RepayAction transaction

    :param token: which token is borrowed
    :type token: str
    :param interest_rate_mode: interest rate mode
    :type interest_rate_mode: InterestRateMode
    :param amount: amount repaid
    :type amount: UnitDecimal
    :param debt_after: total borrow amount of this token after repay transaction
    :type debt_after: UnitDecimal
    """

    token: str
    """which token is borrowed"""
    interest_rate_mode: InterestRateMode
    """interest rate mode"""
    amount: UnitDecimal
    """amount repaid"""
    debt_after: UnitDecimal
    """total borrow amount of this token after repay transaction"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_repay

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.red,
            {
                "token": self.token,
                "interest_rate_mode": self.interest_rate_mode.name,
                "amount": self.amount.to_str(),
                "debt_after": self.debt_after.to_str(),
            },
        )


@dataclass
class LiquidationAction(BaseAction):
    """
    Describe parameters and results of a liquidation

    :param collateral_token: which collateral token is used in liquidation
    :type collateral_token: str
    :param debt_token:  which debt token is used in liquidation
    :type debt_token: str
    :param delt_to_cover: Debt amount to be liquidated, should be equal to variable_delt_liquidated+stable_delt_liquidated
    :type delt_to_cover: UnitDecimal
    :param collateral_used: Collateral amount to subtract in liquidation
    :type collateral_used: UnitDecimal
    :param variable_delt_liquidated: liquidated debt token amount in variable delt
    :type variable_delt_liquidated: UnitDecimal
    :param stable_delt_liquidated: liquidated debt token amount in stable delt
    :type stable_delt_liquidated: UnitDecimal
    :param health_factor_before: health factor before liquidation
    :type health_factor_before: UnitDecimal
    :param health_factor_after: health factor after liquidation
    :type health_factor_after: UnitDecimal
    :param collateral_after: collateral token amount after liquidation
    :type collateral_after: UnitDecimal
    :param variable_debt_after: variable debt token amount after liquidation
    :type variable_debt_after: UnitDecimal
    :param stable_delt_after: stable delt token amount after liquidation
    :type stable_delt_after: UnitDecimal
    """

    collateral_token: str
    """which collateral token is used in liquidation"""
    debt_token: str
    """which debt token is used in liquidation"""
    delt_to_cover: UnitDecimal
    """Debt amount to be liquidated, should be equal to variable_delt_liquidated+stable_delt_liquidated"""
    collateral_used: UnitDecimal
    """Collateral amount to subtract in liquidation"""
    variable_delt_liquidated: UnitDecimal
    """liquidated debt token amount in variable delt"""
    stable_delt_liquidated: UnitDecimal
    """liquidated debt token amount in stable delt"""
    health_factor_before: Decimal
    """health factor before liquidation"""
    health_factor_after: Decimal
    """health factor after liquidation"""
    collateral_after: UnitDecimal
    """collateral token amount after liquidation"""
    variable_debt_after: UnitDecimal
    """variable debt token amount after liquidation"""
    stable_delt_after: UnitDecimal
    """stable delt token amount after liquidation"""

    def set_type(self):
        self.action_type = ActionTypeEnum.aave_repay

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.yellow,
            {
                "collateral_token": self.collateral_token,
                "debt_token": self.debt_token,
                "delt_to_cover": self.delt_to_cover.to_str(),
                "collateral_used": self.collateral_used.to_str(),
                "liquidated": f"variable:{self.variable_delt_liquidated.to_str()} stable:{self.stable_delt_liquidated.to_str()}",
                "health_factor": f"{self.health_factor_before}->{self.health_factor_after}",
                "collateral_after": self.collateral_after.to_str(),
                "variable_debt_after": self.variable_debt_after.to_str(),
                "stable_delt_after": self.stable_delt_after.to_str(),
            },
        )


class DictCache:
    """
    A cache class in dictionary. It uses dict for storage.
    """

    def __init__(self):
        self.empty = True
        self._value: Dict[K, T] = {}

    @property
    def value(self) -> Dict[K, T]:
        """
        Get dict in this cache instance.
        """
        return self._value

    def get(self, k: K) -> T:
        """
        Get value from cache.

        :param k: key
        :type k: K
        :return: value
        :rtype: T
        """
        return self._value[k]

    def reset(self):
        """
        Reset cache instance, will set values to empty.
        """
        self._value: Dict[K, T] = {}
        self.empty = True

    def set(self, k: K, v: T):
        """
        Set value to dict

        :param k: key
        :type k: K
        :param v: value
        :type v: T
        """
        self._value[k] = v
        self.empty = False


@dataclass
class AaveDescription(MarketDescription):
    """
    Designed to generate json description for aave market

    :param type: market type
    :type type: str
    :param name: market name
    :type name: str
    :param supplies_count: count of supplies
    :type supplies_count: int
    :param borrows_count: count of borrows
    :type borrows_count: int
    """

    supplies_count: int
    """count of supplies"""
    borrows_count: int
    """count of borrows"""



import pandas as pd
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import Dict, NamedTuple, Union

from .._typing import TokenInfo, UnitDecimal, MarketDescription
from ..broker import BaseAction, ActionTypeEnum, MarketBalance, MarketStatus
from ..utils import console_text
from ..utils import get_formatted_from_dict
from ..utils.console_text import get_action_str, ForColorEnum


class PositionInfo(NamedTuple):
    """
    get_position information, including tick range and liquidity. It contains immutable properties of a get_position, and used as a key for get_position dict

    :param lower_tick: lower tick
    :type lower_tick: int
    :param upper_tick: upper tick
    :type upper_tick: int

    """

    lower_tick: int
    """lower tick"""
    upper_tick: int
    """upper tick"""

    def __str__(self):
        return f"""tick:{self.lower_tick},{self.upper_tick}"""


@dataclass
class UniDescription(MarketDescription):
    """
    A brief description for uniswap market status.
    """

    token0: TokenInfo
    """token0"""
    token1: TokenInfo
    """token1"""
    quote_token: TokenInfo
    """quote token"""
    base_token: TokenInfo
    """base token"""
    fee_rate: Decimal
    """fee rate"""


class PositionStatus(NamedTuple):
    liquidity: int
    liquidity_amount0: Decimal
    liquidity_amount1: Decimal
    liquidity_value :Decimal
    pending_amount0: Decimal
    pending_amount1: Decimal
    pending_value:Decimal
    amount0: Decimal
    amount1: Decimal
    value: Decimal
    H: Decimal
    L: Decimal
    P: Decimal


class UniV3Pool(object):
    """
    pool information, corresponding with definition in pool contract.

    :param token0: First token in  pool contract.
    :type token0:  TokenInfo
    :param token1: Second token in  pool contract.
    :type token1: TokenInfo
    :param fee: fee rate of this pool, should be among [0.05%, 0.3%, 1%]
    :type fee: float, 0.05
    :param quote_token: which token will be considered as base token. e.g. to a token pair of USDT/BTC, if you want price unit to be like 10000 usdt/btc, you should set usdt as base token, otherwise if price unit is 0.00001 btc/usdt, you should set btc as base token
    :type quote_token: TokenInfo
    """

    def __init__(self, token0: TokenInfo, token1: TokenInfo, fee: float, quote_token: TokenInfo):
        fee = Decimal(str(fee))
        self.token0 = token0
        self.token1 = token1
        self.is_token0_quote = quote_token == token0
        self.quote_token = quote_token
        self.base_token = token1 if self.is_token0_quote else token0
        self.tick_spacing = int(fee * 200)
        self.fee: Decimal = fee * Decimal(10000)
        self.fee_rate: Decimal = Decimal(fee) / Decimal(100)

    def __str__(self):
        return (
            "PoolBaseInfo(Token0: {},".format(self.token0)
            + "Token1: {},".format(self.token1)
            + "fee: {}%,".format(self.fee_rate * Decimal(100))
            + "base token: {})".format(self.token0.name if self.is_token0_quote else self.token1.name)
        )

    def __repr__(self):
        return self.__str__()


@dataclass
class UniLpBalance(MarketBalance):
    """
    current balances of quote and base token in uniswap market

    :param base_uncollected: base token uncollect fee in all the positions.
    :type base_uncollected: UnitDecimal
    :param quote_uncollected: quote token uncollect fee in all the positions.
    :type quote_uncollected: UnitDecimal
    :param base_in_position: base token amount deposited in positions, calculated according to current price
    :type base_in_position: UnitDecimal
    :param quote_in_position: quote token amount deposited in positions, calculated according to current price
    :type quote_in_position: UnitDecimal
    :param position_count: count of positions
    :type position_count: int

    """

    base_uncollected: UnitDecimal
    """base token uncollect fee in all the positions."""
    quote_uncollected: UnitDecimal
    """quote token uncollect fee in all the positions."""
    base_in_position: UnitDecimal
    """base token amount deposited in positions, calculated according to current price"""
    quote_in_position: UnitDecimal
    """quote token amount deposited in positions, calculated according to current price"""
    position_count: int
    """count of positions"""

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """
        return get_formatted_from_dict(
            {
                "total capital": self.net_value.to_str(),
                "uncollect fee": f"{self.base_uncollected.to_str()},{self.quote_uncollected.to_str()}",
                "in get_position amount": f"{self.base_in_position.to_str()},{self.quote_in_position.to_str()}",
                "get_position count": self.position_count,
            }
        )

    def to_array(self):
        """
        Join attributes to an array
        """
        return [
            self.net_value,
            self.base_uncollected,
            self.quote_uncollected,
            self.base_in_position,
            self.quote_in_position,
            self.position_count,
        ]


# @DeprecationWarning
# class BrokerAsset(object):
#     """
#     Wallet of broker, manage balance of an asset.
#     It will prevent excess usage on asset.
#     """
#
#     def __init__(self, token: TokenInfo, init_amount=Decimal(0)):
#         self.token_info = token
#         self.name = token.name
#         self.decimal = token.decimal
#         self.balance = init_amount
#
#     def __str__(self):
#         return f"{self.balance} {self.name}"
#
#     def add(self, amount=Decimal(0)):
#         """
#         add amount to balance
#         :param amount: amount to add
#         :type amount: Decimal
#         :return: entity itself
#         :rtype: BrokerAsset
#         """
#         self.balance += amount
#         return self
#
#     def sub(self, amount=Decimal(0), allow_negative_balance=False):
#         """
#         subtract amount from balance. if balance is not enough, an error will be raised.
#
#         :param amount: amount to subtract
#         :type amount: Decimal
#         :param allow_negative_balance: allow balance is negative
#         :type allow_negative_balance: bool
#         :return:
#         :rtype:
#         """
#         base = self.balance if self.balance != Decimal(0) else Decimal(amount)
#
#         if base == Decimal(0):  # amount and balance is both 0
#             return self
#         if allow_negative_balance:
#             self.balance -= amount
#         else:
#             # if difference between amount and balance is below 0.01%, will deduct all the balance
#             # That's because, the amount calculated by v3_core, has some acceptable error.
#             if abs((self.balance - amount) / base) < 0.00001:
#                 self.balance = Decimal(0)
#             elif self.balance - amount < Decimal(0):
#                 raise DemeterError(f"Insufficient balance, balance is {self.balance}{self.name}, " f"but sub amount is {amount}{self.name}")
#             else:
#                 self.balance -= amount
#
#         return self
#
#     def amount_in_wei(self):
#         return self.balance * Decimal(10**self.decimal)


@dataclass
class Position(object):
    """
    keeps variables for get_position
    """

    pending_amount0: Decimal
    pending_amount1: Decimal
    liquidity: int
    lower_price: Decimal
    upper_price: Decimal
    init_price: Decimal
    transferred: bool = False  # this position(nft) has been transferred, so owner is not current user.


def position_dict_to_dataframe(positions: Dict[PositionInfo, Position]) -> pd.DataFrame:
    pos_dict = {
        "lower_tick": [],
        "upper_tick": [],
        "pending0": [],
        "pending1": [],
        "liquidity": [],
    }
    for k, v in positions.items():
        pos_dict["lower_tick"].append(k.lower_tick)
        pos_dict["upper_tick"].append(k.upper_tick)
        pos_dict["pending0"].append(console_text.format_value(v.pending_amount0))
        pos_dict["pending1"].append(console_text.format_value(v.pending_amount1))
        pos_dict["liquidity"].append(v.liquidity)
    return pd.DataFrame(pos_dict)


@dataclass
class UniV3PoolStatus:
    """
    current status of a pool, actuators can notify current status to broker by filling this entity
    """

    # required by market class
    price: Decimal
    """token price(price of quote token)"""
    currentLiquidity: int = None
    """total liquidity of this pool"""
    inAmount0: int = None
    """in amount of token 0 in this minute"""
    inAmount1: int = None
    """in amount of token 1 in this minute"""
    closeTick: int = None
    """last price tick of this minute"""

    # useless. just pass poperity to
    netAmount0: int = None
    """total amount of token 0"""
    netAmount1: int = None
    """total amount of token 1"""
    openTick: int = None
    """start tick in this minute"""
    lowestTick: int = None
    """lowest tick in this minute"""
    highestTick: int = None
    """highest tick in this minute"""
    open: Decimal = None
    """start price in this minute"""
    low: Decimal = None
    """lowest price in this minute"""
    high: Decimal = None
    """highest price in this minute"""
    volume0: Decimal = None
    """swap volumn of token 0"""
    volume1: Decimal = None
    """swap volumn of token 1"""


@dataclass
class UniswapMarketStatus(MarketStatus):
    """
    MarketStatus properties

    :param data: current pool status
    :type data: Union[pd.Series, UniV3PoolStatus]
    """

    data: Union[pd.Series, UniV3PoolStatus] = None


@dataclass
class UniLpBaseAction(BaseAction):
    """
    Parent class of broker actions,

    :param base_balance_after: after action balance of base token
    :type base_balance_after: UnitDecimal
    :param quote_balance_after: after action balance of quote token
    :type quote_balance_after: UnitDecimal
    """

    base_balance_after: UnitDecimal
    quote_balance_after: UnitDecimal

    def get_output_str(self):
        return str(self)


@dataclass
class AddLiquidityAction(UniLpBaseAction):
    """
    Add Liquidity

    :param base_amount_max: inputted base token amount, also the max amount to deposit
    :type base_amount_max: ActionTypeEnum
    :param quote_amount_max: inputted base token amount, also the max amount to deposit
    :type quote_amount_max: datetime
    :param lower_quote_price: lower price base on quote token.
    :type lower_quote_price: UnitDecimal
    :param upper_quote_price: upper price base on quote token.
    :type upper_quote_price: UnitDecimal
    :param base_amount_actual: actual used base token
    :type base_amount_actual: UnitDecimal
    :param quote_amount_actual: actual used quote token
    :type quote_amount_actual: UnitDecimal
    :param position: generated get_position
    :type position: PositionInfo
    :param liquidity: liquidity added
    :type liquidity: int
    """

    base_amount_max: UnitDecimal
    quote_amount_max: UnitDecimal
    lower_quote_price: UnitDecimal
    upper_quote_price: UnitDecimal
    base_amount_actual: UnitDecimal
    quote_amount_actual: UnitDecimal
    position: PositionInfo
    liquidity: int

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_add_liquidity

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """
        return get_action_str(
            self,
            ForColorEnum.red,
            {
                "max amount": f"{self.base_amount_max.to_str()},{self.quote_amount_max.to_str()}",
                "price": f"{self.lower_quote_price.to_str()},{self.upper_quote_price.to_str()}",
                "get_position": str(self.position),
                "liquidity": self.liquidity,
                "balance": f"{self.base_balance_after.to_str()}(-{self.base_amount_actual.to_str()}), {self.quote_balance_after.to_str()}(-{self.quote_amount_actual.to_str()})",
            },
        )


@dataclass
class CollectFeeAction(UniLpBaseAction):
    """
    collect fee

    :param position: get_position to operate
    :type position: PositionInfo
    :param base_amount: fee collected in base token
    :type base_amount: UnitDecimal
    :param quote_amount: fee collected in quote token
    :type quote_amount: UnitDecimal

    """

    position: PositionInfo
    base_amount: UnitDecimal
    quote_amount: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_collect

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """
        return get_action_str(
            self,
            ForColorEnum.yellow,
            {
                "get_position": str(self.position),
                "balance": f"{self.base_balance_after.to_str()}(+{self.base_amount.to_str()}), {self.quote_balance_after.to_str()}(+{self.quote_amount.to_str()})",
            },
        )


@dataclass
class RemoveLiquidityAction(UniLpBaseAction):
    """
    remove get_position

    :param position: get_position to operate
    :type position: PositionInfo
    :param base_amount: base token amount collected
    :type base_amount: UnitDecimal
    :param quote_amount: quote token amount collected
    :type quote_amount: UnitDecimal
    :param removed_liquidity: liquidity number has removed
    :type removed_liquidity: int
    :param remain_liquidity: liquidity number left in get_position
    :type remain_liquidity: int

    """

    position: PositionInfo
    base_amount: UnitDecimal
    quote_amount: UnitDecimal
    removed_liquidity: int
    remain_liquidity: int

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_remove_liquidity

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.green,
            {
                "get_position": str(self.position),
                "balance": f"{self.base_balance_after.to_str()}(+0), {self.quote_balance_after.to_str()}(+0)",
                "token_got": f"{self.base_amount.to_str()},{self.quote_amount.to_str()}",
                "removed liquidity": self.removed_liquidity,
                "remain liquidity": self.remain_liquidity,
            },
        )


@dataclass
class SwapAction(BaseAction):
    """
    buy token, swap from base token to quote token.

    :param amount: amount to buy(in quote token)
    :type amount: UnitDecimal
    :param price: price,
    :type price: UnitDecimal
    :param fee: fee paid (in base token)
    :type fee: UnitDecimal
    :param base_change: base token amount changed
    :type base_change: PositionInfo
    :param quote_change: quote token amount changed
    :type quote_change: UnitDecimal

    """

    amount: UnitDecimal
    price: UnitDecimal
    fee: UnitDecimal
    to_amount: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_swap

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """
        return get_action_str(
            self,
            ForColorEnum.cyan,
            {
                "price": self.price.to_str(),
                "fee": self.fee.to_str(),
                "amount": f"{self.amount.to_str()}->{self.to_amount.to_str()}",
            },
        )


@dataclass
class BuyAction(UniLpBaseAction):
    """
    buy token, swap from base token to quote token.

    :param amount: amount to buy(in quote token)
    :type amount: UnitDecimal
    :param price: price,
    :type price: UnitDecimal
    :param fee: fee paid (in base token)
    :type fee: UnitDecimal
    :param base_change: base token amount changed
    :type base_change: PositionInfo
    :param quote_change: quote token amount changed
    :type quote_change: UnitDecimal

    """

    amount: UnitDecimal
    price: UnitDecimal
    fee: UnitDecimal
    base_change: UnitDecimal
    quote_change: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_buy

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """
        return get_action_str(
            self,
            ForColorEnum.cyan,
            {
                "price": self.price.to_str(),
                "fee": self.fee.to_str(),
                "balance": f"{self.base_balance_after.to_str()}(-{self.base_change.to_str()}), {self.quote_balance_after.to_str()}(+{self.quote_change.to_str()})",
            },
        )


@dataclass
class SellAction(UniLpBaseAction):
    """
    sell token, swap from quote token to base token.

    :param amount: amount to sell(in quote token)
    :type amount: UnitDecimal
    :param price: price,
    :type price: UnitDecimal
    :param fee: fee paid (in quote token)
    :type fee: UnitDecimal
    :param base_change: base token amount changed
    :type base_change: PositionInfo
    :param quote_change: quote token amount changed
    :type quote_change: UnitDecimal

    """

    amount: UnitDecimal
    price: UnitDecimal
    fee: UnitDecimal
    base_change: UnitDecimal
    quote_change: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_sell

    def get_output_str(self):
        """
        get colored and formatted string to output in console

        :return: formatted string
        :rtype: str
        """

        return get_action_str(
            self,
            ForColorEnum.light_red,
            {
                "price": self.price.to_str(),
                "fee": self.fee.to_str(),
                "balance": f"{self.base_balance_after.to_str()}(+{self.base_change.to_str()}), {self.quote_balance_after.to_str()}(-{self.quote_change.to_str()})",
            },
        )
